import shutil
from datetime import datetime
from pathlib import Path
from typing import Any, Optional, TypeVar

import tomli
import tomli_w

from flood_adapt.dbs_classes.interface.database import IDatabase
from flood_adapt.dbs_classes.interface.element import AbstractDatabaseElement
from flood_adapt.objects.object_model import Object

T_OBJECTMODEL = TypeVar("T_OBJECTMODEL", bound=Object)


class DbsTemplate(AbstractDatabaseElement[T_OBJECTMODEL]):
    display_name: str
    dir_name: str
    _object_class: type[T_OBJECTMODEL]

    def __init__(
        self, database: IDatabase, standard_objects: Optional[list[str]] = None
    ):
        """Initialize any necessary attributes."""
        self._database = database
        self.input_path = database.input_path / self.dir_name
        self.output_path = database.output_path / self.dir_name
        self.standard_objects = standard_objects

    def get(self, name: str) -> T_OBJECTMODEL:
        """Return an object of the type of the database with the given name.

        Parameters
        ----------
        name : str
            name of the object to be returned

        Returns
        -------
        Object
            object of the type of the specified object model
        """
        # Make the full path to the object
        full_path = self.input_path / name / f"{name}.toml"

        # Check if the object exists
        if not Path(full_path).is_file():
            raise ValueError(f"{self.display_name}: '{name}' does not exist.")

        # Load and return the object
        return self._object_class.load_file(full_path)

    def summarize_objects(self) -> dict[str, list[Any]]:
        """Return a dictionary with info on the objects that currently exist in the database.

        Returns
        -------
        dict[str, list[Any]]
            A dictionary that contains the keys: `name`, `description`, `path`  and `last_modification_date`.
            Each key has a list of the corresponding values, where the index of the values corresponds to the same object.
        """
        return self._get_object_summary()

    def copy(self, old_name: str, new_name: str, new_description: str):
        """Copy (duplicate) an existing object, and give it a new name.

        Parameters
        ----------
        old_name : str
            name of the existing measure
        new_name : str
            name of the new measure
        new_description : str
            description of the new measure
        """
        copy_object = self.get(old_name)
        copy_object.name = new_name
        copy_object.description = new_description

        # After changing the name and description, re-trigger the validators
        copy_object.model_validate(copy_object)

        # Checking whether the new name is already in use
        self._validate_to_save(copy_object, overwrite=False)

        # Write only the toml file
        toml_path = self.input_path / new_name / f"{new_name}.toml"
        toml_path.parent.mkdir(parents=True)
        with open(toml_path, "wb") as f:
            tomli_w.dump(copy_object.model_dump(exclude_none=True), f)

        # Then copy all the accompanied files
        src = self.input_path / old_name
        dest = self.input_path / new_name

        EXCLUDE = [".spw", ".toml"]
        for file in src.glob("*"):
            if file.suffix in EXCLUDE:
                continue
            if file.is_dir():
                shutil.copytree(file, dest / file.name, dirs_exist_ok=True)
            else:
                shutil.copy2(file, dest / file.name)

    def save(
        self,
        object_model: T_OBJECTMODEL,
        overwrite: bool = False,
    ):
        """Save an object in the database and all associated files.

        This saves the toml file and any additional files attached to the object.

        Parameters
        ----------
        object_model : Object
            object to be saved in the database
        overwrite : bool, optional
            whether to overwrite the object if it already exists in the
            database, by default False

        Raises
        ------
        ValueError
            Raise error if name is already in use.
        """
        self._validate_to_save(object_model, overwrite=overwrite)

        # If the folder doesnt exist yet, make the folder and save the object
        if not (self.input_path / object_model.name).exists():
            (self.input_path / object_model.name).mkdir()

        # Save the object and any additional files
        object_model.save(
            self.input_path / object_model.name / f"{object_model.name}.toml",
        )

    def delete(self, name: str, toml_only: bool = False):
        """Delete an already existing object in the database.

        Parameters
        ----------
        name : str
            name of the object to be deleted
        toml_only : bool, optional
            whether to only delete the toml file or the entire folder. If the folder is empty after deleting the toml,
            it will always be deleted. By default False

        Raises
        ------
        ValueError
            Raise error if object to be deleted is already in use.
        """
        # Check if the object is a standard object. If it is, raise an error
        if self._check_standard_objects(name):
            raise ValueError(
                f"'{name}' cannot be deleted/modified since it is a standard {self.display_name}."
            )

        # Check if object is used in a higher level object. If it is, raise an error
        if used_in := self.check_higher_level_usage(name):
            raise ValueError(
                f"{self.display_name}: '{name}' cannot be deleted/modified since it is already used in: {', '.join(used_in)}"
            )

        # Once all checks are passed, delete the object
        toml_path = self.input_path / name / f"{name}.toml"
        if toml_only:
            # Only delete the toml file
            toml_path.unlink(missing_ok=True)
            # If the folder is empty, delete the folder
            if not list(toml_path.parent.iterdir()):
                toml_path.parent.rmdir()
        else:
            # Delete the entire folder
            shutil.rmtree(toml_path.parent, ignore_errors=True)
            if (self.output_path / name).exists():
                shutil.rmtree(self.output_path / name, ignore_errors=True)

    def _check_standard_objects(self, name: str) -> bool:
        """Check if an object is a standard object.

        Parameters
        ----------
        name : str
            name of the object to be checked

        Returns
        -------
        bool
            True if the object is a standard object, False otherwise
        """
        if self.standard_objects:
            return name in self.standard_objects
        return False

    def check_higher_level_usage(self, name: str) -> list[str]:
        """Check if an object is used in a higher level object.

        Parameters
        ----------
        name : str
            name of the object to be checked

        Returns
        -------
        list[str]
            list of higher level objects that use the object
        """
        # If this function is not implemented for the object type, it cannot be used in a higher
        # level object. By default, return an empty list
        return []

    def _get_object_summary(self) -> dict[str, list[Any]]:
        """Get a dictionary with all the toml paths and last modification dates that exist in the database of the given object_type.

        Returns
        -------
        dict[str, Any]
            A dictionary that contains the keys: `name`, `description`, `path`  and `last_modification_date`.
            Each key has a list of the corresponding values, where the index of the values corresponds to the same object.
        """
        # If the toml doesnt exist, we might be in the middle of saving a new object or could be a broken object.
        # In any case, we should not list it in the database
        directories = [
            dir
            for dir in self.input_path.iterdir()
            if (dir / f"{dir.name}.toml").is_file()
        ]
        paths = [Path(dir / f"{dir.name}.toml") for dir in directories]

        names = [self._read_variable_in_toml("name", path) for path in paths]
        descriptions = [
            self._read_variable_in_toml("description", path) for path in paths
        ]

        last_modification_date = [
            datetime.fromtimestamp(file.stat().st_mtime) for file in paths
        ]

        objects = {
            "name": names,
            "description": descriptions,
            "path": paths,
            "last_modification_date": last_modification_date,
        }
        return objects

    @staticmethod
    def _read_variable_in_toml(variable_name: str, toml_path: Path) -> str:
        with open(toml_path, "rb") as f:
            data = tomli.load(f)
        return data.get(variable_name, "")

    def _validate_to_save(self, object_model: T_OBJECTMODEL, overwrite: bool) -> None:
        """Validate if the object can be saved.

        Parameters
        ----------
        object_model : Object
            object to be validated

        Raises
        ------
        ValueError
            Raise error if name is already in use.
        """
        # Check if the object exists
        object_exists = object_model.name in self.summarize_objects()["name"]

        # If you want to overwrite the object, and the object already exists, first delete it. If it exists and you
        # don't want to overwrite, raise an error.
        if overwrite and object_exists:
            self.delete(object_model.name, toml_only=True)
        elif not overwrite and object_exists:
            raise ValueError(
                f"'{object_model.name}' name is already used by another {self.display_name.lower()}. Choose a different name"
            )



import os
from enum import Enum
from pathlib import Path
from typing import Any, ClassVar, List, Optional, Protocol, runtime_checkable

import tomli
from pydantic import (
    Field,
    field_serializer,
    field_validator,
    model_validator,
)

from flood_adapt.config.config import Settings
from flood_adapt.objects.forcing.forcing import (
    ForcingSource,
    ForcingType,
    IForcing,
)
from flood_adapt.objects.forcing.forcing_factory import ForcingFactory
from flood_adapt.objects.forcing.time_frame import TimeFrame
from flood_adapt.objects.object_model import Object


class Mode(str, Enum):
    """Class describing the accepted input for the variable mode in Event.

    Attributes
    ----------
    single_event : The single event mode.
    risk : The risk mode.
    """

    single_event = "single_event"
    risk = "risk"


class Template(str, Enum):
    """Class describing the accepted input for the variable template in Event.

    Attributes
    ----------
    Synthetic : The synthetic template.
    Hurricane : The hurricane template.
    Historical : The historical template.
    """

    Synthetic = "Synthetic"
    Hurricane = "Hurricane"
    Historical = "Historical"

    @property
    def description(self) -> str:
        match self:
            case Template.Historical:
                return "Select a time period for a historic event. This method can use offshore wind and pressure fields for the selected time period to simulate nearshore water levels or download gauged waterlevels to perform a realistic simulation. These water levels are used together with rainfall and river discharge input to simulate flooding in the site area."
            case Template.Hurricane:
                return "Select a historical hurricane track from the hurricane database, and shift the track if desired."
            case Template.Synthetic:
                return "Customize a synthetic event by specifying the waterlevels, wind, rainfall and river discharges without being based on a historical event."
            case _:
                raise ValueError(f"Invalid event template: {self}")


@runtime_checkable
class PathBasedForcing(Protocol):
    """Protocol for forcing classes that have a path attribute.

    Performing an isinstance check on this class will return True if the class has a path attribute (even if it is None).
    """

    path: Path


class Event(Object):
    """The accepted input for an event in FloodAdapt.

    Attributes
    ----------
    name : str
        The name of the event.
    description : str
        The description of the event. Defaults to "".
    time : TimeFrame
        The time frame of the event.
    template : Template
        The template of the event.
    mode : Mode
        The mode of the event.
    rainfall_multiplier : float
        The rainfall multiplier of the event.
    forcings : dict[ForcingType, list[IForcing]]
        The forcings of the event.
    """

    ALLOWED_FORCINGS: ClassVar[dict[ForcingType, List[ForcingSource]]]

    time: TimeFrame
    template: Template
    mode: Mode = Mode.single_event

    forcings: dict[ForcingType, list[IForcing]] = Field(default_factory=dict)
    rainfall_multiplier: float = Field(default=1.0, ge=0)

    @classmethod
    def get_allowed_forcings(cls) -> dict[str, List[str]]:
        return {k.value: [s.value for s in v] for k, v in cls.ALLOWED_FORCINGS.items()}

    def get_forcings(self) -> list[IForcing]:
        """Return a list of all forcings in the event."""
        return [forcing for forcings in self.forcings.values() for forcing in forcings]

    def save_additional(self, output_dir: Path | str | os.PathLike) -> None:
        """Save any additional files associated with the event."""
        for forcing in self.get_forcings():
            forcing.save_additional(output_dir)

    @classmethod
    def load_file(cls, file_path: Path | str | os.PathLike) -> "Event":
        """Load object from file.

        Parameters
        ----------
        file_path : Path | str | os.PathLike
            Path to the file to load.

        """
        with open(file_path, mode="rb") as fp:
            toml = tomli.load(fp)

        event = cls.model_validate(toml)

        # Update all forcings with paths to absolute paths
        for forcing in event.get_forcings():
            if isinstance(forcing, PathBasedForcing):
                if forcing.path.exists():
                    continue
                elif forcing.path == Path(forcing.path.name):
                    # convert relative path to absolute path
                    in_dir = Path(file_path).parent / forcing.path.name
                    if not in_dir.exists():
                        raise FileNotFoundError(
                            f"Failed to load Event. File {forcing.path} does not exist in {in_dir.parent}."
                        )
                    forcing.path = in_dir
                else:
                    raise FileNotFoundError(
                        f"Failed to load Event. File {forcing.path} does not exist."
                    )

        return event

    @staticmethod
    def _parse_forcing_from_dict(
        forcing_attrs: dict[str, Any] | IForcing,
        ftype: Optional[ForcingType] = None,
        fsource: Optional[ForcingSource] = None,
    ) -> IForcing:
        if isinstance(forcing_attrs, IForcing):
            # forcing_attrs is already a forcing object
            return forcing_attrs
        elif isinstance(forcing_attrs, dict):
            # forcing_attrs is a dict with valid forcing attributes
            if "type" not in forcing_attrs and ftype:
                forcing_attrs["type"] = ftype
            if "source" not in forcing_attrs and fsource:
                forcing_attrs["source"] = fsource

            return ForcingFactory.load_dict(forcing_attrs)
        else:
            raise ValueError(
                f"Invalid forcing attributes: {forcing_attrs}. "
                "Forcings must be one of:\n"
                "1. Instance of IForcing\n"
                "2. dict with the keys `type` (ForcingType), `source` (ForcingSource) specifying the class, and with valid forcing attributes for that class."
            )

    @field_validator("forcings", mode="before")
    @classmethod
    def create_forcings(
        cls, value: dict[str, list[dict[str, Any]]]
    ) -> dict[ForcingType, list[IForcing]]:
        forcings = {}
        for ftype, forcing_list in value.items():
            ftype = ForcingType(ftype)
            forcings[ftype] = [
                cls._parse_forcing_from_dict(forcing, ftype) for forcing in forcing_list
            ]
        return forcings

    @model_validator(mode="after")
    def validate_forcings(self):
        def validate_concrete_forcing(concrete_forcing):
            type = concrete_forcing.type
            source = concrete_forcing.source

            # Check type
            if type not in self.__class__.ALLOWED_FORCINGS:
                allowed_types = ", ".join(
                    t.value for t in self.__class__.ALLOWED_FORCINGS.keys()
                )
                raise ValueError(
                    f"Forcing type {type.value} is not allowed. Allowed types are: {allowed_types}"
                )

            # Check source
            if source not in self.__class__.ALLOWED_FORCINGS[type]:
                allowed_sources = ", ".join(
                    s.value for s in self.__class__.ALLOWED_FORCINGS[type]
                )
                raise ValueError(
                    f"Forcing source {source.value} is not allowed for forcing type {type.value}. "
                    f"Allowed sources are: {allowed_sources}"
                )

        if Settings().validate_allowed_forcings and hasattr(self, "ALLOWED_FORCINGS"):
            # Validate forcings
            for _, concrete_forcings in self.forcings.items():
                for concrete_forcing in concrete_forcings:
                    validate_concrete_forcing(concrete_forcing)

        return self

    @field_serializer("forcings")
    @classmethod
    def serialize_forcings(
        cls, value: dict[ForcingType, List[IForcing]]
    ) -> dict[str, List[dict[str, Any]]]:
        dct = {}
        for ftype, forcing_list in value.items():
            dct[ftype.value] = [
                forcing.model_dump(exclude_none=True) for forcing in forcing_list
            ]
        return dct



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Union

import geopandas as gpd
import pandas as pd
from cht_cyclones.cyclone_track_database import CycloneTrackDatabase

from flood_adapt.adapter.sfincs_adapter import SfincsAdapter


class IDbsStatic(ABC):
    @abstractmethod
    def get_aggregation_areas(self) -> dict: ...

    @abstractmethod
    def get_model_boundary(self) -> gpd.GeoDataFrame: ...

    @abstractmethod
    def get_model_grid(self): ...

    @abstractmethod
    def get_obs_points(self) -> gpd.GeoDataFrame: ...

    @abstractmethod
    def get_static_map(self, path: Union[str, Path]) -> gpd.GeoDataFrame: ...

    @abstractmethod
    def get_slr_scn_names(self) -> list: ...

    @abstractmethod
    def get_green_infra_table(self, measure_type: str) -> pd.DataFrame: ...

    @abstractmethod
    def get_buildings(self) -> gpd.GeoDataFrame: ...

    @abstractmethod
    def get_property_types(self) -> list: ...

    @abstractmethod
    def get_overland_sfincs_model(self) -> SfincsAdapter: ...

    @abstractmethod
    def get_offshore_sfincs_model(self) -> SfincsAdapter: ...

    @abstractmethod
    def get_cyclone_track_database(self) -> CycloneTrackDatabase: ...



import os
from pathlib import Path
from typing import List, Optional

import tomli
from pydantic import BaseModel

from flood_adapt.objects.events.events import Event, Mode
from flood_adapt.objects.object_model import Object


class SubEventModel(BaseModel):
    """The accepted input for a sub event in FloodAdapt.

    Attributes
    ----------
    name : str
        The name of the sub event.
    frequency : float
        The frequency of the sub event.
    """

    name: str
    frequency: float


class EventSet(Object):
    """BaseModel describing the expected variables and data types for parameters of EventSet.

    An EventSet is a collection of events that can be used to create a scenario and perform a probabilistoc risk assessment.

    Attributes
    ----------
    name : str
        The name of the event.
    description : str
        The description of the event. Defaults to "".
    mode : Mode
        The mode of the event. Defaults to Mode.risk.
    sub_events : List[SubEventModel]
        The sub events of the event set.
    """

    _events: Optional[List[Event]] = None

    mode: Mode = Mode.risk
    sub_events: List[SubEventModel]

    def load_sub_events(
        self,
        sub_events: Optional[List[Event]] = None,
        file_path: Optional[Path] = None,
    ) -> None:
        """Load sub events from a list or from a file path."""
        if sub_events is not None:
            self._events = sub_events
        elif file_path is not None:
            from flood_adapt.objects.events.event_factory import EventFactory

            sub_events = []
            for sub_event in self.sub_events:
                sub_toml = (
                    Path(file_path).parent / sub_event.name / f"{sub_event.name}.toml"
                )
                sub_events.append(EventFactory.load_file(sub_toml))

            self._events = sub_events
        else:
            raise ValueError("Either `sub_events` or `file_path` must be provided.")

    @classmethod
    def load_file(cls, file_path: Path | str | os.PathLike):
        """Load object from file."""
        with open(file_path, mode="rb") as fp:
            event_set = tomli.load(fp)
        event_set = EventSet(**event_set)
        event_set.load_sub_events(file_path=file_path)
        return event_set

    def save_additional(self, output_dir: Path | str | os.PathLike) -> None:
        for sub_event in self._events:
            sub_dir = Path(output_dir) / sub_event.name
            sub_dir.mkdir(parents=True, exist_ok=True)
            sub_event.save(sub_dir / f"{sub_event.name}.toml")



import math
import os
from pathlib import Path
from typing import Annotated

import numpy as np
import pandas as pd
from pydantic import BaseModel

from flood_adapt.misc.utils import (
    copy_file_to_output_dir,
    validate_file_extension,
)
from flood_adapt.objects.forcing import unit_system as us
from flood_adapt.objects.forcing.forcing import (
    ForcingSource,
    IWaterlevel,
)
from flood_adapt.objects.forcing.time_frame import (
    TimeFrame,
)
from flood_adapt.objects.forcing.timeseries import (
    CSVTimeseries,
    SyntheticTimeseries,
    TimeseriesFactory,
)


class SurgeModel(BaseModel):
    """BaseModel describing the expected variables and data types for harmonic tide parameters of synthetic model."""

    timeseries: SyntheticTimeseries


class TideModel(BaseModel):
    """BaseModel describing the expected variables and data types for harmonic tide parameters of synthetic model."""

    harmonic_amplitude: us.UnitfulLength
    harmonic_phase: us.UnitfulTime
    harmonic_period: us.UnitfulTime = us.UnitfulTime(
        value=12.4, units=us.UnitTypesTime.hours
    )

    def to_dataframe(self, time_frame: TimeFrame) -> pd.DataFrame:
        index = pd.date_range(
            start=time_frame.start_time,
            end=time_frame.end_time,
            freq=time_frame.time_step,
            name="time",
        )
        seconds = np.arange(len(index)) * time_frame.time_step.total_seconds()
        seconds.round(decimals=0)

        amp = self.harmonic_amplitude.value
        omega = 2 * math.pi / (self.harmonic_period.convert(us.UnitTypesTime.seconds))
        phase_seconds = self.harmonic_phase.convert(us.UnitTypesTime.seconds)

        tide = amp * np.cos(omega * (seconds - phase_seconds))
        return pd.DataFrame(data=tide, index=index)


class WaterlevelSynthetic(IWaterlevel):
    source: ForcingSource = ForcingSource.SYNTHETIC

    surge: SurgeModel
    tide: TideModel

    def to_dataframe(self, time_frame: TimeFrame) -> pd.DataFrame:
        tide_df = self.tide.to_dataframe(time_frame=time_frame)

        surge_df = TimeseriesFactory.from_object(self.surge.timeseries).to_dataframe(
            time_frame=time_frame
        )

        # Combine
        tide_df.columns = ["waterlevel"]
        surge_df.columns = ["waterlevel"]
        surge_df = surge_df.reindex(tide_df.index, method="nearest", limit=1).fillna(
            value=self.surge.timeseries.fill_value
        )

        wl_df = tide_df.add(surge_df, axis="index", fill_value=0)
        wl_df.columns = ["waterlevel"]

        return wl_df


class WaterlevelCSV(IWaterlevel):
    source: ForcingSource = ForcingSource.CSV

    path: Annotated[Path, validate_file_extension([".csv"])]
    units: us.UnitTypesLength = us.UnitTypesLength.meters

    def to_dataframe(self, time_frame: TimeFrame) -> pd.DataFrame:
        return CSVTimeseries.load_file(
            path=self.path, units=us.UnitfulLength(value=0, units=self.units)
        ).to_dataframe(time_frame=time_frame)

    def save_additional(self, output_dir: Path | str | os.PathLike) -> None:
        self.path = copy_file_to_output_dir(self.path, Path(output_dir))


class WaterlevelModel(IWaterlevel):
    source: ForcingSource = ForcingSource.MODEL


class WaterlevelGauged(IWaterlevel):
    source: ForcingSource = ForcingSource.GAUGED



import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Optional, Union

import geopandas as gpd
import numpy as np
import pandas as pd
from cht_cyclones.tropical_cyclone import TropicalCyclone

from flood_adapt.config.site import Site
from flood_adapt.dbs_classes.interface.element import AbstractDatabaseElement
from flood_adapt.dbs_classes.interface.static import IDbsStatic
from flood_adapt.objects.benefits.benefits import Benefit
from flood_adapt.objects.events.events import Event


class IDatabase(ABC):
    base_path: Path
    input_path: Path
    output_path: Path
    static_path: Path

    @property
    @abstractmethod
    def site(self) -> Site: ...

    @property
    @abstractmethod
    def static(self) -> IDbsStatic: ...

    @property
    @abstractmethod
    def events(self) -> AbstractDatabaseElement: ...

    @property
    @abstractmethod
    def scenarios(self) -> AbstractDatabaseElement: ...

    @property
    @abstractmethod
    def strategies(self) -> AbstractDatabaseElement: ...

    @property
    @abstractmethod
    def measures(self) -> AbstractDatabaseElement: ...

    @property
    @abstractmethod
    def projections(self) -> AbstractDatabaseElement: ...

    @property
    @abstractmethod
    def benefits(self) -> AbstractDatabaseElement: ...

    @abstractmethod
    def __init__(
        self, database_path: Union[str, os.PathLike], site_name: str
    ) -> None: ...

    @abstractmethod
    def interp_slr(self, slr_scenario: str, year: float) -> float:
        pass

    @abstractmethod
    def plot_slr_scenarios(self) -> str:
        pass

    @abstractmethod
    def write_to_csv(self, name: str, event: Event, df: pd.DataFrame) -> None:
        pass

    @abstractmethod
    def write_cyc(self, event: Event, track: TropicalCyclone) -> None:
        pass

    @abstractmethod
    def check_benefit_scenarios(self, benefit: Benefit) -> pd.DataFrame:
        pass

    @abstractmethod
    def create_benefit_scenarios(self, benefit: Benefit) -> None:
        pass

    @abstractmethod
    def run_benefit(self, benefit_name: Union[str, list[str]]) -> None:
        pass

    @abstractmethod
    def get_outputs(self) -> dict[str, Any]:
        pass

    @abstractmethod
    def get_topobathy_path(self) -> str:
        pass

    @abstractmethod
    def get_index_path(self) -> str:
        pass

    @abstractmethod
    def get_depth_conversion(self) -> float:
        pass

    @abstractmethod
    def get_max_water_level(
        self, scenario_name: str, return_period: Optional[int] = None
    ) -> np.ndarray:
        pass

    @abstractmethod
    def get_building_footprints(self, scenario_name: str) -> gpd.GeoDataFrame:
        pass

    @abstractmethod
    def get_roads(self, scenario_name: str) -> gpd.GeoDataFrame:
        pass

    @abstractmethod
    def get_aggregation(self, scenario_name: str) -> dict[str, gpd.GeoDataFrame]:
        pass

    @abstractmethod
    def get_aggregation_benefits(
        self, benefit_name: str
    ) -> dict[str, gpd.GeoDataFrame]:
        pass

    @abstractmethod
    def get_object_list(self, object_type: str) -> dict[str, Any]:
        pass

    @abstractmethod
    def has_run_hazard(self, scenario_name: str) -> None:
        pass

    @abstractmethod
    def cleanup(self) -> None:
        pass



import os
from typing import Union

import tomli

from flood_adapt.objects.measures.measures import (
    Buyout,
    Elevate,
    FloodProof,
    FloodWall,
    GreenInfrastructure,
    Measure,
    MeasureType,
    Pump,
)


class MeasureFactory:
    @staticmethod
    def get_measure_type(filepath: Union[str, os.PathLike]):
        """Get a measure type from toml file."""
        with open(filepath, mode="rb") as fp:
            toml = tomli.load(fp)
        type = toml.get("type")
        return MeasureType(type)

    @staticmethod
    def get_measure_object(filepath: Union[str, os.PathLike]) -> Measure:
        measure_type = MeasureFactory.get_measure_type(filepath)

        if MeasureType.is_impact(measure_type):
            return ImpactMeasureFactory.get_impact_measure(measure_type).load_file(
                filepath
            )
        elif MeasureType.is_hazard(measure_type):
            return HazardMeasureFactory.get_hazard_measure(measure_type).load_file(
                filepath
            )
        else:
            raise ValueError(f"Measure type {measure_type} not recognized.")


class ImpactMeasureFactory:
    """Simple parser to get the respective measure subclass from a measure type string given in the config file.

    Args:
        type (str): name of measure type

    Returns
    -------
        Measure: Measure subclass
    """

    @staticmethod
    def get_impact_measure(impact_measure: str):
        if impact_measure == "elevate_properties":
            return Elevate
        elif impact_measure == "buyout_properties":
            return Buyout
        elif impact_measure == "floodproof_properties":
            return FloodProof
        else:
            raise ValueError(f"Measure type {impact_measure} not recognized.")


class HazardMeasureFactory:
    """Simple parser to get the respective measure subclass from a measure type string given in the config file.

    Args:
        type (str): name of measure type

    Returns
    -------
        Measure: Measure subclass
    """

    @staticmethod
    def get_hazard_measure(hazard_measure: str):
        if (
            hazard_measure == "floodwall" or hazard_measure == "levee"
        ):  # these should use the same class right?
            return FloodWall
        elif (
            hazard_measure == "water_square"
            or hazard_measure == "total_storage"
            or hazard_measure == "greening"
        ):
            return GreenInfrastructure
        elif hazard_measure == "pump":
            return Pump
        else:
            raise ValueError(f"Measure type {hazard_measure} not recognized.")



import shutil

from flood_adapt.dbs_classes.dbs_template import DbsTemplate
from flood_adapt.workflows.benefit_runner import Benefit, BenefitRunner


class DbsBenefit(DbsTemplate[Benefit]):
    display_name = "Benefit"
    dir_name = "benefits"
    _object_class = Benefit

    def save(self, object_model: Benefit, overwrite: bool = False):
        """Save a benefit object in the database.

        Parameters
        ----------
        object_model : Benefit
            object of Benefit type
        overwrite : bool, optional
            whether to overwrite existing benefit with same name, by default False

        Raises
        ------
        ValueError
            Raise error if name is already in use. Names of benefits assessments should be unique.
        """
        runner = BenefitRunner(self._database, benefit=object_model)

        # Check if all scenarios are created
        if not all(runner.scenarios["scenario created"] != "No"):
            raise ValueError(
                f"'{object_model.name}' name cannot be created before all necessary scenarios are created."
            )

        # Save the benefit
        super().save(object_model, overwrite=overwrite)

    def delete(self, name: str, toml_only: bool = False):
        """Delete an already existing benefit in the database.

        Parameters
        ----------
        name : str
            name of the benefit
        toml_only : bool, optional
            whether to only delete the toml file or the entire folder. If the folder is empty after deleting the toml,
            it will always be deleted. By default False

        Raises
        ------
        ValueError
            Raise error if benefit has already model output
        """
        # First delete the benefit
        super().delete(name, toml_only=toml_only)

        # Delete output if edited
        output_path = self.output_path / name
        if output_path.exists():
            shutil.rmtree(output_path, ignore_errors=True)

    def get_runner(self, name: str) -> BenefitRunner:
        return BenefitRunner(self._database, self.get(name))

    def has_run_check(self, name: str) -> bool:
        return self.get_runner(name).has_run_check()

    def ready_to_run(self, name: str) -> bool:
        """Check if all the required scenarios have already been run.

        Returns
        -------
        bool
            True if required scenarios have been already run
        """
        return self.get_runner(name).ready_to_run()



import logging
import math
import os
import shutil
import subprocess
import tempfile
from pathlib import Path
from typing import Optional, Union

import geopandas as gpd
import hydromt_sfincs.utils as utils
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import pyproj
import shapely
import xarray as xr
from cht_cyclones.tropical_cyclone import TropicalCyclone
from cht_tide.read_bca import SfincsBoundary
from cht_tide.tide_predict import predict
from hydromt_sfincs import SfincsModel as HydromtSfincsModel
from hydromt_sfincs.quadtree import QuadtreeGrid
from numpy import matlib
from shapely.affinity import translate

from flood_adapt.adapter.interface.hazard_adapter import IHazardAdapter
from flood_adapt.config.config import Settings
from flood_adapt.config.site import Site
from flood_adapt.misc.log import FloodAdaptLogging
from flood_adapt.misc.path_builder import (
    ObjectDir,
    TopLevelDir,
    db_path,
)
from flood_adapt.misc.utils import cd, resolve_filepath
from flood_adapt.objects.events.event_set import EventSet
from flood_adapt.objects.events.events import Event, Mode, Template
from flood_adapt.objects.events.historical import HistoricalEvent
from flood_adapt.objects.events.hurricane import TranslationModel
from flood_adapt.objects.forcing import unit_system as us
from flood_adapt.objects.forcing.discharge import (
    DischargeConstant,
    DischargeCSV,
    DischargeSynthetic,
)
from flood_adapt.objects.forcing.forcing import (
    ForcingSource,
    ForcingType,
    IDischarge,
    IForcing,
    IRainfall,
    IWaterlevel,
    IWind,
)
from flood_adapt.objects.forcing.meteo_handler import MeteoHandler
from flood_adapt.objects.forcing.rainfall import (
    RainfallConstant,
    RainfallCSV,
    RainfallMeteo,
    RainfallNetCDF,
    RainfallSynthetic,
    RainfallTrack,
)
from flood_adapt.objects.forcing.time_frame import TimeFrame
from flood_adapt.objects.forcing.waterlevels import (
    WaterlevelCSV,
    WaterlevelGauged,
    WaterlevelModel,
    WaterlevelSynthetic,
)
from flood_adapt.objects.forcing.wind import (
    WindConstant,
    WindCSV,
    WindMeteo,
    WindNetCDF,
    WindSynthetic,
    WindTrack,
)
from flood_adapt.objects.measures.measures import (
    FloodWall,
    GreenInfrastructure,
    Measure,
    Pump,
)
from flood_adapt.objects.projections.projections import (
    PhysicalProjection,
    Projection,
)
from flood_adapt.objects.scenarios.scenarios import Scenario


class SfincsAdapter(IHazardAdapter):
    """Adapter for the SFINCS model.

    This class is used to run the SFINCS model and process the results.

    Attributes
    ----------
    settings : SfincsModel
        The settings for the SFINCS model.
    """

    logger = FloodAdaptLogging.getLogger("SfincsAdapter")
    _site: Site
    _model: HydromtSfincsModel

    ###############
    ### PUBLIC ####
    ###############

    ### HAZARD ADAPTER METHODS ###
    def __init__(self, model_root: Path):
        """Load overland sfincs model based on a root directory.

        Parameters
        ----------
        model_root : Path
            Root directory of overland sfincs model.
        """
        self.settings = self.database.site.sfincs
        self.units = self.database.site.gui.units
        self.sfincs_logger = self._setup_sfincs_logger(model_root)
        self._model = HydromtSfincsModel(
            root=str(model_root.resolve()), mode="r", logger=self.sfincs_logger
        )
        self._model.read()

    def read(self, path: Path):
        """Read the sfincs model from the current model root."""
        if Path(self._model.root).resolve() != Path(path).resolve():
            self._model.set_root(root=str(path), mode="r")
        self._model.read()

    def write(self, path_out: Union[str, os.PathLike], overwrite: bool = True):
        """Write the sfincs model configuration to a directory."""
        root = self.get_model_root()
        if not isinstance(path_out, Path):
            path_out = Path(path_out).resolve()

        if not path_out.exists():
            path_out.mkdir(parents=True)

        if root != path_out:
            shutil.copytree(root, path_out, dirs_exist_ok=True)

        write_mode = "w+" if overwrite else "w"
        with cd(path_out):
            self._model.set_root(root=str(path_out), mode=write_mode)
            self._model.write()

    def close_files(self):
        """Close all open files and clean up file handles."""
        for logger in [self.logger, self.sfincs_logger]:
            if hasattr(logger, "handlers"):
                for handler in logger.handlers:
                    if isinstance(handler, logging.FileHandler):
                        handler.close()
                        logger.removeHandler(handler)

    def __enter__(self) -> "SfincsAdapter":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        self.close_files()
        return False

    def has_run(self, scenario: Scenario) -> bool:
        """Check if the model has been run."""
        event = self.database.events.get(scenario.event)
        if event.mode == Mode.risk:
            sim_paths = [
                self._get_simulation_path(scenario, sub_event=sub_event)
                for sub_event in event.sub_events
            ]
            # No need to check postprocessing for risk scenarios
            return all(self.sfincs_completed(sim_path) for sim_path in sim_paths)
        else:
            return self.sfincs_completed(self._get_simulation_path(scenario))

    def execute(self, path: Path, strict: bool = True) -> bool:
        """
        Run the sfincs executable in the specified path.

        Parameters
        ----------
        path : str
            Path to the simulation folder.
            Default is None, in which case the model root is used.
        strict : bool, optional
            True: raise an error if the model fails to run.
            False: log a warning.
            Default is True.

        Returns
        -------
        bool
            True if the model ran successfully, False otherwise.

        """
        with cd(path):
            self.logger.info(f"Running SFINCS in {path}")
            process = subprocess.run(
                str(Settings().sfincs_path),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            self.sfincs_logger.info(process.stdout)
            self.logger.debug(process.stdout)

        self._cleanup_simulation_folder(path)

        if process.returncode != 0:
            if Settings().delete_crashed_runs:
                # Remove all files in the simulation folder except for the log files
                for subdir, dirs, files in os.walk(path, topdown=False):
                    for file in files:
                        if not file.endswith(".log"):
                            os.remove(os.path.join(subdir, file))

                    if not os.listdir(subdir):
                        os.rmdir(subdir)

            if strict:
                raise RuntimeError(f"SFINCS model failed to run in {path}.")
            else:
                self.logger.error(f"SFINCS model failed to run in {path}.")

        return process.returncode == 0

    def run(self, scenario: Scenario):
        """Run the whole workflow (Preprocess, process and postprocess) for a given scenario."""
        self._ensure_no_existing_forcings()
        event = self.database.events.get(scenario.event)

        if event.mode == Mode.risk:
            self._run_risk_scenario(scenario=scenario)
        else:
            self._run_single_event(scenario=scenario, event=event)

    def preprocess(self, scenario: Scenario, event: Event):
        """
        Preprocess the SFINCS model for a given scenario.

        Parameters
        ----------
        scenario : Scenario
            Scenario to preprocess.
        event : Event, optional
            Event to preprocess, by default None.
        """
        # I dont like this due to it being state based and might break if people use functions in the wrong order
        # Currently only used to pass projection + event stuff to WaterlevelModel

        sim_path = self._get_simulation_path(scenario=scenario, sub_event=event)
        sim_path.mkdir(parents=True, exist_ok=True)
        template_path = (
            self.database.static.get_overland_sfincs_model().get_model_root()
        )
        shutil.copytree(template_path, sim_path, dirs_exist_ok=True)

        with SfincsAdapter(model_root=sim_path) as model:
            model._load_scenario_objects(scenario, event)
            is_risk = "Probabilistic " if model._event_set is not None else ""
            self.logger.info(
                f"Preprocessing Scenario `{model._scenario.name}`: {is_risk}Event `{model._event.name}`, Strategy `{model._strategy.name}`, Projection `{model._projection.name}`"
            )
            # Write template model to output path and set it as the model root so focings can write to it
            model.set_timing(model._event.time)
            model.write(sim_path)

            # Event
            for forcing in model._event.get_forcings():
                model.add_forcing(forcing)

            if self.rainfall is not None:
                model.rainfall *= model._event.rainfall_multiplier
            else:
                model.logger.warning(
                    "Failed to add event rainfall multiplier, no rainfall forcing found in the model."
                )

            # Measures
            for measure in model._strategy.get_hazard_measures():
                model.add_measure(measure)

            # Projection
            model.add_projection(model._projection)

            # Output
            model.add_obs_points()

            # Save any changes made to disk as well
            model.write(path_out=sim_path)

    def process(self, scenario: Scenario, event: Event):
        if event.mode != Mode.single_event:
            raise ValueError(f"Unsupported event mode: {event.mode}.")

        sim_path = self._get_simulation_path(scenario=scenario, sub_event=event)
        self.logger.info(f"Running SFINCS for single event Scenario `{scenario.name}`")
        self.execute(sim_path)

    def postprocess(self, scenario: Scenario, event: Event):
        if event.mode != Mode.single_event:
            raise ValueError(f"Unsupported event mode: {event.mode}.")

        self.logger.info(f"Postprocessing SFINCS for Scenario `{scenario.name}`")
        if not self.sfincs_completed(
            self._get_simulation_path(scenario, sub_event=event)
        ):
            raise RuntimeError("SFINCS was not run successfully!")

        self.write_floodmap_geotiff(scenario)
        self.plot_wl_obs(scenario)
        self.write_water_level_map(scenario)

    def set_timing(self, time: TimeFrame):
        """Set model reference times."""
        self.logger.info(f"Setting timing for the SFINCS model: `{time}`")
        self._model.set_config("tref", time.start_time)
        self._model.set_config("tstart", time.start_time)
        self._model.set_config("tstop", time.end_time)

    def add_forcing(self, forcing: IForcing):
        """Get forcing data and add it."""
        if forcing is None:
            return

        self.logger.info(
            f"Adding {forcing.type.capitalize()}: {forcing.source.capitalize()}"
        )
        if isinstance(forcing, IRainfall):
            self._add_forcing_rain(forcing)
        elif isinstance(forcing, IWind):
            self._add_forcing_wind(forcing)
        elif isinstance(forcing, IDischarge):
            self._add_forcing_discharge(forcing)
        elif isinstance(forcing, IWaterlevel):
            self._add_forcing_waterlevels(forcing)
        else:
            self.logger.warning(
                f"Skipping unsupported forcing type {forcing.__class__.__name__}"
            )

    def add_measure(self, measure: Measure):
        """Get measure data and add it."""
        self.logger.info(
            f"Adding {measure.__class__.__name__.capitalize()} `{measure.name}`"
        )

        if isinstance(measure, FloodWall):
            self._add_measure_floodwall(measure)
        elif isinstance(measure, GreenInfrastructure):
            self._add_measure_greeninfra(measure)
        elif isinstance(measure, Pump):
            self._add_measure_pump(measure)
        else:
            self.logger.warning(
                f"Skipping unsupported measure type {measure.__class__.__name__}"
            )

    def add_projection(self, projection: Projection):
        """Get forcing data currently in the sfincs model and add the projection it."""
        self.logger.info(f"Adding Projection `{projection.name}`")
        phys_projection = projection.physical_projection

        if phys_projection.sea_level_rise:
            self.logger.info(
                f"Adding projected sea level rise `{phys_projection.sea_level_rise}`"
            )
            if self.waterlevels is not None:
                self.waterlevels += phys_projection.sea_level_rise.convert(
                    us.UnitTypesLength.meters
                )
            else:
                self.logger.warning(
                    "Failed to add sea level rise, no water level forcing found in the model."
                )

        if phys_projection.rainfall_multiplier:
            self.logger.info(
                f"Adding projected rainfall multiplier `{phys_projection.rainfall_multiplier}`"
            )
            if self.rainfall is not None:
                self.rainfall *= phys_projection.rainfall_multiplier
            else:
                self.logger.warning(
                    "Failed to add projected rainfall multiplier, no rainfall forcing found in the model."
                )

    ### GETTERS ###
    def get_model_time(self) -> TimeFrame:
        t0, t1 = self._model.get_model_time()
        return TimeFrame(start_time=t0, end_time=t1)

    def get_model_root(self) -> Path:
        return Path(self._model.root)

    def get_mask(self):
        """Get mask with inactive cells from model."""
        mask = self._model.grid["msk"]
        return mask

    def get_bedlevel(self):
        """Get bed level from model."""
        self._model.read_results()
        zb = self._model.results["zb"]
        return zb

    def get_model_boundary(self) -> gpd.GeoDataFrame:
        """Get bounding box from model."""
        return self._model.region

    def get_model_grid(self) -> QuadtreeGrid:
        """Get grid from model.

        Returns
        -------
        QuadtreeGrid
            QuadtreeGrid with the model grid
        """
        return self._model.quadtree

    # Forcing properties
    @property
    def waterlevels(self) -> xr.Dataset | xr.DataArray | None:
        return self._model.forcing.get("bzs")

    @waterlevels.setter
    def waterlevels(self, waterlevels: xr.Dataset | xr.DataArray):
        if self.waterlevels is None or self.waterlevels.size == 0:
            raise ValueError("No water level forcing found in the model.")
        self._model.forcing["bzs"] = waterlevels

    @property
    def discharge(self) -> xr.Dataset | xr.DataArray | None:
        return self._model.forcing.get("dis")

    @discharge.setter
    def discharge(self, discharge: xr.Dataset | xr.DataArray):
        if self.discharge is None or self.discharge.size == 0:
            raise ValueError("No discharge forcing found in the model.")
        self._model.forcing["dis"] = discharge

    @property
    def rainfall(self) -> xr.Dataset | xr.DataArray | None:
        names = ["precip", "precip_2d"]
        in_model = [name for name in names if name in self._model.forcing]
        if len(in_model) == 0:
            return None
        elif len(in_model) == 1:
            return self._model.forcing[in_model[0]]
        else:
            raise ValueError("Multiple rainfall forcings found in the model.")

    @rainfall.setter
    def rainfall(self, rainfall: xr.Dataset | xr.DataArray):
        if self.rainfall is None or self.rainfall.size == 0:
            raise ValueError("No rainfall forcing found in the model.")
        elif "precip_2d" in self._model.forcing:
            self._model.forcing["precip_2d"] = rainfall
        elif "precip" in self._model.forcing:
            self._model.forcing["precip"] = rainfall
        else:
            raise ValueError("Unsupported rainfall forcing in the model.")

    @property
    def wind(self) -> xr.Dataset | xr.DataArray | None:
        wind_names = ["wnd", "wind_2d", "wind", "wind10_u", "wind10_v"]
        wind_in_model = [name for name in wind_names if name in self._model.forcing]
        if len(wind_in_model) == 0:
            return None
        elif len(wind_in_model) == 1:
            return self._model.forcing[wind_in_model[0]]
        elif len(wind_in_model) == 2:
            if not ("wind10_u" in wind_in_model and "wind10_v" in wind_in_model):
                raise ValueError(
                    "Multiple wind forcings found in the model. Both should be wind10_u and wind10_v or a singular wind forcing."
                )
            return xr.Dataset(
                {
                    "wind10_u": self._model.forcing["wind10_u"],
                    "wind10_v": self._model.forcing["wind10_v"],
                }
            )
        else:
            raise ValueError("Multiple wind forcings found in the model.")

    @wind.setter
    def wind(self, wind: xr.Dataset | xr.DataArray):
        if (not self.wind) or (self.wind.size == 0):
            raise ValueError("No wind forcing found in the model.")

        elif "wind_2d" in self._model.forcing:
            self._model.forcing["wind_2d"] = wind
        elif "wind" in self._model.forcing:
            self._model.forcing["wind"] = wind
        elif "wnd" in self._model.forcing:
            self._model.forcing["wnd"] = wind
        elif "wind10_u" in self._model.forcing and "wind10_v" in self._model.forcing:
            self._model.forcing["wind10_u"] = wind["wind10_u"]
            self._model.forcing["wind10_v"] = wind["wind10_v"]
        else:
            raise ValueError("Unsupported wind forcing in the model.")

    ### OUTPUT ###
    def run_completed(self, scenario: Scenario) -> bool:
        """Check if the entire model run has been completed successfully by checking if all flood maps exist that are created in postprocess().

        Returns
        -------
        bool : True if all flood maps exist, False otherwise.

        """
        any_floodmap = len(self._get_flood_map_paths(scenario)) > 0
        all_exist = all(
            floodmap.exists() for floodmap in self._get_flood_map_paths(scenario)
        )
        return any_floodmap and all_exist

    def sfincs_completed(self, sim_path: Path) -> bool:
        """Check if the sfincs executable has been run successfully by checking if the output files exist in the simulation folder.

        Parameters
        ----------
        sim_path : Path
            Path to the simulation folder to check.

        Returns
        -------
        bool: True if the sfincs executable has been run successfully, False otherwise.

        """
        SFINCS_OUTPUT_FILES = ["sfincs_map.nc"]

        if self.settings.obs_point is not None:
            SFINCS_OUTPUT_FILES.append("sfincs_his.nc")

        to_check = [Path(sim_path) / file for file in SFINCS_OUTPUT_FILES]
        return all(output.exists() for output in to_check)

    def write_floodmap_geotiff(
        self, scenario: Scenario, sim_path: Optional[Path] = None
    ):
        """
        Read simulation results from SFINCS and saves a geotiff with the maximum water levels.

        Produced floodmap is in the units defined in the sfincs config settings.

        Parameters
        ----------
        scenario : Scenario
            Scenario for which to create the floodmap.
        sim_path : Path, optional
            Path to the simulation folder, by default None.
        """
        self.logger.info("Writing flood maps to geotiff")
        results_path = self._get_result_path(scenario)
        sim_path = sim_path or self._get_simulation_path(scenario)
        demfile = self.database.static_path / "dem" / self.settings.dem.filename

        with SfincsAdapter(model_root=sim_path) as model:
            zsmax = model._get_zsmax()

            dem = model._model.data_catalog.get_rasterdataset(demfile)

            # convert dem from dem units to floodmap units
            dem_conversion = us.UnitfulLength(
                value=1.0, units=self.settings.dem.units
            ).convert(self.settings.config.floodmap_units)

            floodmap_fn = results_path / f"FloodMap_{scenario.name}.tif"

            # convert zsmax from meters to floodmap units
            floodmap_conversion = us.UnitfulLength(
                value=1.0, units=us.UnitTypesLength.meters
            ).convert(self.settings.config.floodmap_units)

            utils.downscale_floodmap(
                zsmax=floodmap_conversion * zsmax,
                dep=dem_conversion * dem,
                hmin=0.01,
                floodmap_fn=str(floodmap_fn),
            )

    def write_water_level_map(
        self, scenario: Scenario, sim_path: Optional[Path] = None
    ):
        """Read simulation results from SFINCS and saves a netcdf with the maximum water levels."""
        self.logger.info("Writing water level map to netcdf")
        results_path = self._get_result_path(scenario)
        sim_path = sim_path or self._get_simulation_path(scenario)

        with SfincsAdapter(model_root=sim_path) as model:
            zsmax = model._get_zsmax()
            zsmax.to_netcdf(results_path / "max_water_level_map.nc")

    def plot_wl_obs(
        self,
        scenario: Scenario,
    ):
        """Plot water levels at SFINCS observation points as html.

        Only for single event scenarios, or for a specific simulation path containing the written and processed sfincs model.
        """
        if not self.settings.obs_point:
            self.logger.warning("No observation points provided in config.")
            return

        self.logger.info("Plotting water levels at observation points")
        sim_path = self._get_simulation_path(scenario)

        # read SFINCS model
        with SfincsAdapter(model_root=sim_path) as model:
            df, gdf = model._get_zs_points()

        gui_units = us.UnitTypesLength(
            self.database.site.gui.units.default_length_units
        )
        conversion_factor = us.UnitfulLength(
            value=1.0, units=us.UnitTypesLength("meters")
        ).convert(gui_units)

        overland_reference_height = self.settings.water_level.get_datum(
            self.settings.config.overland_model.reference
        ).height.convert(gui_units)

        for ii, col in enumerate(df.columns):
            # Plot actual thing
            fig = px.line(
                df[col] * conversion_factor
                + overland_reference_height  # convert to reference datum for plotting
            )

            fig.add_hline(
                y=0,
                line_dash="dash",
                line_color="#000000",
                annotation_text=self.settings.water_level.reference,
                annotation_position="bottom right",
            )

            # plot reference water levels
            for wl_ref in self.settings.water_level.datums:
                if (
                    wl_ref.name == self.settings.config.overland_model.reference
                    or wl_ref.name in self.database.site.gui.plotting.excluded_datums
                ):
                    continue
                fig.add_hline(
                    y=wl_ref.height.convert(gui_units),
                    line_dash="dash",
                    line_color="#3ec97c",
                    annotation_text=wl_ref.name,
                    annotation_position="bottom right",
                )

            fig.update_layout(
                autosize=False,
                height=100 * 2,
                width=280 * 2,
                margin={"r": 0, "l": 0, "b": 0, "t": 20},
                font={"size": 10, "color": "black", "family": "Arial"},
                title={
                    "text": gdf.iloc[ii]["Description"],
                    "font": {"size": 12, "color": "black", "family": "Arial"},
                    "x": 0.5,
                    "xanchor": "center",
                },
                xaxis_title="Time",
                yaxis_title=f"Water level [{gui_units.value}] above {self.settings.water_level.reference}",
                yaxis_title_font={"size": 10, "color": "black", "family": "Arial"},
                xaxis_title_font={"size": 10, "color": "black", "family": "Arial"},
                showlegend=False,
            )

            event = self.database.events.get(scenario.event)
            self._add_tide_gauge_plot(fig, event, units=gui_units)

            # write html to results folder
            station_name = gdf.iloc[ii]["Name"]
            results_path = self._get_result_path(scenario)
            fig.write_html(results_path / f"{station_name}_timeseries.html")

    def add_obs_points(self):
        """Add observation points provided in the site toml to SFINCS model."""
        if self.settings.obs_point is not None:
            self.logger.info("Adding observation points to the overland flood model")

            obs_points = self.settings.obs_point
            names = []
            lat = []
            lon = []
            for pt in obs_points:
                names.append(pt.name)
                lat.append(pt.lat)
                lon.append(pt.lon)

            # create GeoDataFrame from obs_points in site file
            df = pd.DataFrame({"name": names})
            gdf = gpd.GeoDataFrame(
                df, geometry=gpd.points_from_xy(lon, lat), crs="EPSG:4326"
            )

            # Add locations to SFINCS file
            self._model.setup_observation_points(locations=gdf, merge=False)

    def get_wl_df_from_offshore_his_results(self) -> pd.DataFrame:
        """Create a pd.Dataframe with waterlevels from the offshore model at the bnd locations of the overland model.

        Returns
        -------
        wl_df: pd.DataFrame
            time series of water level.
        """
        self.logger.info("Reading water levels from offshore model")
        ds_his = utils.read_sfincs_his_results(
            Path(self._model.root) / "sfincs_his.nc",
            crs=self._model.crs.to_epsg(),
        )
        wl_df = pd.DataFrame(
            data=ds_his.point_zs.to_numpy(),
            index=ds_his.time.to_numpy(),
            columns=np.arange(1, ds_his.point_zs.to_numpy().shape[1] + 1, 1),
        )
        return wl_df

    ## RISK EVENTS ##
    def calculate_rp_floodmaps(self, scenario: Scenario):
        """Calculate flood risk maps from a set of (currently) SFINCS water level outputs using linear interpolation.

        It would be nice to make it more widely applicable and move the loading of the SFINCS results to self.postprocess_sfincs().

        generates return period water level maps in netcdf format to be used by FIAT
        generates return period water depth maps in geotiff format as product for users

        TODO: make this robust and more efficient for bigger datasets.
        """
        event: EventSet = self.database.events.get(scenario.event)
        if not isinstance(event, EventSet):
            raise ValueError("This function is only available for risk scenarios.")

        result_path = self._get_result_path(scenario)
        sim_paths = [
            self._get_simulation_path(scenario, sub_event=sub_event)
            for sub_event in event._events
        ]

        phys_proj = self.database.projections.get(
            scenario.projection
        ).physical_projection

        floodmap_rp = self.database.site.fiat.risk.return_periods
        frequencies = [sub_event.frequency for sub_event in event.sub_events]

        # adjust storm frequency for hurricane events
        if not math.isclose(phys_proj.storm_frequency_increase, 0, abs_tol=1e-9):
            storminess_increase = phys_proj.storm_frequency_increase / 100.0
            for ii, event in enumerate(event._events):
                if event.template == Template.Hurricane:
                    frequencies[ii] = frequencies[ii] * (1 + storminess_increase)

        with SfincsAdapter(model_root=sim_paths[0]) as dummymodel:
            # read mask and bed level
            mask = dummymodel.get_mask().stack(z=("x", "y"))
            zb = dummymodel.get_bedlevel().stack(z=("x", "y")).to_numpy()

        zs_maps = []
        for simulation_path in sim_paths:
            # read zsmax data from overland sfincs model
            with SfincsAdapter(model_root=simulation_path) as sim:
                zsmax = sim._get_zsmax().load()
                zs_stacked = zsmax.stack(z=("x", "y"))
                zs_maps.append(zs_stacked)

        # Create RP flood maps

        # 1a: make a table of all water levels and associated frequencies
        zs = xr.concat(zs_maps, pd.Index(frequencies, name="frequency"))
        # Get the indices of columns with all NaN values
        nan_cells = np.where(np.all(np.isnan(zs), axis=0))[0]
        # fill nan values with minimum bed levels in each grid cell, np.interp cannot ignore nan values
        zs = xr.where(np.isnan(zs), np.tile(zb, (zs.shape[0], 1)), zs)
        # Get table of frequencies
        freq = np.tile(frequencies, (zs.shape[1], 1)).transpose()

        # 1b: sort water levels in descending order and include the frequencies in the sorting process
        # (i.e. each h-value should be linked to the same p-values as in step 1a)
        sort_index = zs.argsort(axis=0)
        sorted_prob = np.flipud(np.take_along_axis(freq, sort_index, axis=0))
        sorted_zs = np.flipud(np.take_along_axis(zs.values, sort_index, axis=0))

        # 1c: Compute exceedance probabilities of water depths
        # Method: accumulate probabilities from top to bottom
        prob_exceed = np.cumsum(sorted_prob, axis=0)

        # 1d: Compute return periods of water depths
        # Method: simply take the inverse of the exceedance probability (1/Pex)
        rp_zs = 1.0 / prob_exceed

        # For each return period (T) of interest do the following:
        # For each grid cell do the following:
        # Use the table from step [1d] as a “lookup-table” to derive the T-year water depth. Use a 1-d interpolation technique:
        # h(T) = interp1 (log(T*), h*, log(T))
        # in which t* and h* are the values from the table and T is the return period (T) of interest
        # The resulting T-year water depths for all grids combined form the T-year hazard map
        rp_da = xr.DataArray(rp_zs, dims=zs.dims)

        # no_data_value = -999  # in SFINCS
        # sorted_zs = xr.where(sorted_zs == no_data_value, np.nan, sorted_zs)

        valid_cells = np.where(mask == 1)[
            0
        ]  # only loop over cells where model is not masked
        h = matlib.repmat(
            np.copy(zb), len(floodmap_rp), 1
        )  # if not flooded (i.e. not in valid_cells) revert to bed_level, read from SFINCS results so it is the minimum bed level in a grid cell

        self.logger.info("Calculating flood risk maps, this may take some time")
        for jj in valid_cells:  # looping over all non-masked cells.
            # linear interpolation for all return periods to evaluate
            h[:, jj] = np.interp(
                np.log10(floodmap_rp),
                np.log10(rp_da[::-1, jj]),
                sorted_zs[::-1, jj],
                left=0,
            )

        # Re-fill locations that had nan water level for all simulations with nans
        h[:, nan_cells] = np.full(h[:, nan_cells].shape, np.nan)

        # If a cell has the same water-level as the bed elevation it should be dry (turn to nan)
        diff = h - np.tile(zb, (h.shape[0], 1))
        dry = (
            diff < 10e-10
        )  # here we use a small number instead of zero for rounding errors
        h[dry] = np.nan

        for ii, rp in enumerate(floodmap_rp):
            # #create single nc
            zs_rp_single = xr.DataArray(
                data=h[ii, :], coords={"z": zs["z"]}, attrs={"units": "meters"}
            ).unstack()
            zs_rp_single = zs_rp_single.rio.write_crs(
                zsmax.raster.crs
            )  # , inplace=True)
            zs_rp_single = zs_rp_single.to_dataset(name="risk_map")
            fn_rp = result_path / f"RP_{rp:04d}_maps.nc"
            zs_rp_single.to_netcdf(fn_rp)

            # write geotiff
            # dem file for high resolution flood depth map
            demfile = self.database.static_path / "dem" / self.settings.dem.filename

            # writing the geotiff to the scenario results folder
            with SfincsAdapter(model_root=sim_paths[0]) as dummymodel:
                dem = dummymodel._model.data_catalog.get_rasterdataset(demfile)
                zsmax = zs_rp_single.to_array().squeeze().transpose()
                floodmap_fn = fn_rp.with_suffix(".tif")

                # convert dem from dem units to floodmap units
                dem_conversion = us.UnitfulLength(
                    value=1.0, units=self.settings.dem.units
                ).convert(self.settings.config.floodmap_units)

                # convert zsmax from meters to floodmap units
                floodmap_conversion = us.UnitfulLength(
                    value=1.0, units=us.UnitTypesLength.meters
                ).convert(self.settings.config.floodmap_units)

                utils.downscale_floodmap(
                    zsmax=floodmap_conversion * zsmax,
                    dep=dem_conversion * dem,
                    hmin=0.01,
                    floodmap_fn=str(floodmap_fn),
                )

    ######################################
    ### PRIVATE - use at your own risk ###
    ######################################
    def _run_single_event(self, scenario: Scenario, event: Event):
        self.preprocess(scenario, event)
        self.process(scenario, event)
        self.postprocess(scenario, event)
        shutil.rmtree(
            self._get_simulation_path(scenario, sub_event=event), ignore_errors=True
        )

    def _run_risk_scenario(self, scenario: Scenario):
        """Run the whole workflow for a risk scenario.

        This means preprocessing and running the SFINCS model for each event in the event set, and then postprocessing the results.
        """
        event_set: EventSet = self.database.events.get(scenario.event)
        total = len(event_set._events)

        for i, sub_event in enumerate(event_set._events):
            sim_path = self._get_simulation_path(scenario, sub_event=sub_event)

            # Preprocess
            self.preprocess(scenario, event=sub_event)
            self.logger.info(
                f"Running SFINCS for Eventset Scenario `{scenario.name}`, Event `{sub_event.name}` ({i + 1}/{total})"
            )
            self.execute(sim_path)

        # Postprocess
        self.calculate_rp_floodmaps(scenario)

        # Cleanup
        for i, sub_event in enumerate(event_set._events):
            shutil.rmtree(
                self._get_simulation_path(scenario, sub_event=sub_event),
                ignore_errors=True,
            )

    def _ensure_no_existing_forcings(self):
        """Check for existing forcings in the model and raise an error if any are found."""
        all_forcings = {
            "waterlevel": self.waterlevels,
            "rainfall": self.rainfall,
            "wind": self.wind,
            "discharge": self.discharge,
        }
        contains_forcings = ", ".join(
            [
                f"{name.capitalize()}"
                for name, forcing in all_forcings.items()
                if forcing is not None
            ]
        )
        if contains_forcings:
            raise ValueError(
                f"{contains_forcings} forcing(s) should not exists in the SFINCS template model. Remove it from the SFINCS model located at: {self.get_model_root()}. For more information on SFINCS and its input files, see the SFINCS documentation at: `https://sfincs.readthedocs.io/en/latest/input.html`"
            )

    ### FORCING ###
    def _add_forcing_wind(
        self,
        wind: IWind,
    ):
        """Add spatially constant wind forcing to sfincs model. Use timeseries or a constant magnitude and direction.

        Parameters
        ----------
        timeseries : Union[str, os.PathLike], optional
            path to file of timeseries file (.csv) which has three columns: time, magnitude and direction, by default None
        const_mag : float, optional
            magnitude of time-invariant wind forcing [m/s], by default None
        const_dir : float, optional
            direction of time-invariant wind forcing [deg], by default None
        """
        time_frame = self.get_model_time()
        if isinstance(wind, WindConstant):
            # HydroMT function: set wind forcing from constant magnitude and direction
            self._model.setup_wind_forcing(
                timeseries=None,
                magnitude=wind.speed.convert(us.UnitTypesVelocity.mps),
                direction=wind.direction.value,
            )
        elif isinstance(wind, WindSynthetic):
            df = wind.to_dataframe(time_frame=time_frame)
            df["mag"] *= us.UnitfulVelocity(
                value=1.0, units=self.units.default_velocity_units
            ).convert(us.UnitTypesVelocity.mps)

            tmp_path = Path(tempfile.gettempdir()) / "wind.csv"
            df.to_csv(tmp_path)

            # HydroMT function: set wind forcing from timeseries
            self._model.setup_wind_forcing(
                timeseries=tmp_path, magnitude=None, direction=None
            )
        elif isinstance(wind, WindMeteo):
            ds = MeteoHandler().read(time_frame)
            # data already in metric units so no conversion needed

            # HydroMT function: set wind forcing from grid
            self._model.setup_wind_forcing_from_grid(wind=ds)
        elif isinstance(wind, WindTrack):
            # data already in metric units so no conversion needed
            self._add_forcing_spw(wind)
        elif isinstance(wind, WindNetCDF):
            ds = wind.read()
            # time slicing to time_frame not needed, hydromt-sfincs handles it
            conversion = us.UnitfulVelocity(value=1.0, units=wind.units).convert(
                us.UnitTypesVelocity.mps
            )
            ds *= conversion
            self._model.setup_wind_forcing_from_grid(wind=ds)
        elif isinstance(wind, WindCSV):
            df = wind.to_dataframe(time_frame=time_frame)

            conversion = us.UnitfulVelocity(
                value=1.0, units=wind.units["speed"]
            ).convert(us.UnitTypesVelocity.mps)
            df *= conversion

            tmp_path = Path(tempfile.gettempdir()) / "wind.csv"
            df.to_csv(tmp_path)

            # HydroMT function: set wind forcing from timeseries
            self._model.setup_wind_forcing(
                timeseries=tmp_path,
                magnitude=None,
                direction=None,
            )
        else:
            self.logger.warning(
                f"Unsupported wind forcing type: {wind.__class__.__name__}"
            )
            return

    def _add_forcing_rain(self, rainfall: IRainfall):
        """Add spatially constant rain forcing to sfincs model. Use timeseries or a constant magnitude.

        Parameters
        ----------
        timeseries : Union[str, os.PathLike], optional
            path to file of timeseries file (.csv) which has two columns: time and precipitation, by default None
        const_intensity : float, optional
            time-invariant precipitation intensity [mm_hr], by default None
        """
        time_frame = self.get_model_time()
        if isinstance(rainfall, RainfallConstant):
            self._model.setup_precip_forcing(
                timeseries=None,
                magnitude=rainfall.intensity.convert(us.UnitTypesIntensity.mm_hr),
            )
        elif isinstance(rainfall, RainfallCSV):
            df = rainfall.to_dataframe(time_frame=time_frame)
            conversion = us.UnitfulIntensity(value=1.0, units=rainfall.units).convert(
                us.UnitTypesIntensity.mm_hr
            )
            df *= conversion

            tmp_path = Path(tempfile.gettempdir()) / "precip.csv"
            df.to_csv(tmp_path)

            self._model.setup_precip_forcing(timeseries=tmp_path)
        elif isinstance(rainfall, RainfallSynthetic):
            df = rainfall.to_dataframe(time_frame=time_frame)

            if rainfall.timeseries.cumulative is not None:  # scs
                conversion = us.UnitfulLength(
                    value=1.0, units=rainfall.timeseries.cumulative.units
                ).convert(us.UnitTypesLength.millimeters)
            else:
                conversion = us.UnitfulIntensity(
                    value=1.0, units=rainfall.timeseries.peak_value.units
                ).convert(us.UnitTypesIntensity.mm_hr)

            df *= conversion
            tmp_path = Path(tempfile.gettempdir()) / "precip.csv"
            df.to_csv(tmp_path)

            self._model.setup_precip_forcing(timeseries=tmp_path)
        elif isinstance(rainfall, RainfallMeteo):
            ds = MeteoHandler().read(time_frame)
            # MeteoHandler always return metric so no conversion needed
            self._model.setup_precip_forcing_from_grid(precip=ds, aggregate=False)
        elif isinstance(rainfall, RainfallTrack):
            # data already in metric units so no conversion needed
            self._add_forcing_spw(rainfall)
        elif isinstance(rainfall, RainfallNetCDF):
            ds = rainfall.read()
            # time slicing to time_frame not needed, hydromt-sfincs handles it
            conversion = us.UnitfulIntensity(value=1.0, units=rainfall.units).convert(
                us.UnitTypesIntensity.mm_hr
            )
            ds *= conversion
            self._model.setup_precip_forcing_from_grid(precip=ds, aggregate=False)
        else:
            self.logger.warning(
                f"Unsupported rainfall forcing type: {rainfall.__class__.__name__}"
            )
            return

    def _add_forcing_discharge(self, forcing: IDischarge):
        """Add spatially constant discharge forcing to sfincs model. Use timeseries or a constant magnitude.

        Parameters
        ----------
        forcing : IDischarge
            The discharge forcing to add to the model.
            Can be a constant, synthetic or from a csv file.
            Also contains the river information.
        """
        if isinstance(forcing, (DischargeConstant, DischargeCSV, DischargeSynthetic)):
            self._set_single_river_forcing(discharge=forcing)
        else:
            self.logger.warning(
                f"Unsupported discharge forcing type: {forcing.__class__.__name__}"
            )

    def _add_forcing_waterlevels(self, forcing: IWaterlevel):
        time_frame = self.get_model_time()
        if isinstance(forcing, WaterlevelSynthetic):
            df_ts = forcing.to_dataframe(time_frame=time_frame)

            conversion = us.UnitfulLength(
                value=1.0, units=forcing.surge.timeseries.peak_value.units
            ).convert(us.UnitTypesLength.meters)
            datum_correction = self.settings.water_level.get_datum(
                self.database.site.gui.plotting.synthetic_tide.datum
            ).height.convert(us.UnitTypesLength.meters)

            df_ts = df_ts * conversion + datum_correction

            self._set_waterlevel_forcing(df_ts)
        elif isinstance(forcing, WaterlevelGauged):
            if self.settings.tide_gauge is None:
                raise ValueError("No tide gauge defined for this site.")

            df_ts = self.settings.tide_gauge.get_waterlevels_in_time_frame(
                time=time_frame,
            )
            conversion = us.UnitfulLength(
                value=1.0, units=self.settings.tide_gauge.units
            ).convert(us.UnitTypesLength.meters)

            datum_height = self.settings.water_level.get_datum(
                self.settings.tide_gauge.reference
            ).height.convert(us.UnitTypesLength.meters)

            df_ts = conversion * df_ts + datum_height

            self._set_waterlevel_forcing(df_ts)
        elif isinstance(forcing, WaterlevelCSV):
            df_ts = forcing.to_dataframe(time_frame=time_frame)

            if df_ts is None:
                raise ValueError("Failed to get waterlevel data.")
            conversion = us.UnitfulLength(value=1.0, units=forcing.units).convert(
                us.UnitTypesLength.meters
            )
            df_ts *= conversion
            self._set_waterlevel_forcing(df_ts)

        elif isinstance(forcing, WaterlevelModel):
            from flood_adapt.adapter.sfincs_offshore import OffshoreSfincsHandler

            if self.settings.config.offshore_model is None:
                raise ValueError("Offshore model configuration is missing.")
            if self._scenario is None or self._event is None:
                raise ValueError(
                    "Scenario and event must be provided to run the offshore model."
                )

            df_ts = OffshoreSfincsHandler(
                scenario=self._scenario, event=self._event
            ).get_resulting_waterlevels()
            if df_ts is None:
                raise ValueError("Failed to get waterlevel data.")

            # Datum
            datum_correction = self.settings.water_level.get_datum(
                self.settings.config.offshore_model.reference
            ).height.convert(us.UnitTypesLength.meters)
            df_ts += datum_correction

            # Already in meters since it was produced by SFINCS so no conversion needed
            self._set_waterlevel_forcing(df_ts)
            self._turn_off_bnd_press_correction()
        else:
            self.logger.warning(
                f"Unsupported waterlevel forcing type: {forcing.__class__.__name__}"
            )

    # SPIDERWEB
    def _add_forcing_spw(self, forcing: Union[RainfallTrack, WindTrack]):
        """Add spiderweb forcing."""
        if forcing.source != ForcingSource.TRACK:
            raise ValueError("Forcing source should be TRACK.")

        if forcing.path is None:
            raise ValueError("No path to track file provided.")

        if not forcing.path.exists():
            # Check if the file is in the database
            in_db = self._get_event_input_path(self._event) / forcing.path.name
            if not in_db.exists():
                raise FileNotFoundError(
                    f"Input file for track forcing not found: {forcing.path}"
                )
            forcing.path = in_db

        if forcing.path.suffix == ".cyc":
            forcing.path = self._create_spw_file_from_track(
                track_forcing=forcing,
                hurricane_translation=self._event.hurricane_translation,
                name=self._event.name,
                output_dir=forcing.path.parent,
                include_rainfall=bool(self._event.forcings.get(ForcingType.RAINFALL)),
                recreate=False,
            )

        if forcing.path.suffix != ".spw":
            raise ValueError(
                "Track files should be in one of [spw, ddb_cyc] file format and must have [.spw, .cyc] extension."
            )

        sim_path = self.get_model_root()
        self.logger.info(f"Adding spiderweb forcing to Sfincs model: {sim_path.name}")

        # prevent SameFileError
        output_spw_path = sim_path / forcing.path.name
        if forcing.path == output_spw_path:
            raise ValueError(
                "Add a different SPW file than the one already in the model."
            )

        if output_spw_path.exists():
            os.remove(output_spw_path)
        shutil.copy2(forcing.path, output_spw_path)

        self._model.set_config("spwfile", output_spw_path.name)

    ### MEASURES ###
    def _add_measure_floodwall(self, floodwall: FloodWall):
        """Add floodwall to sfincs model.

        Parameters
        ----------
        floodwall : FloodWall
            floodwall information
        """
        polygon_file = resolve_filepath(
            object_dir=ObjectDir.measure,
            obj_name=floodwall.name,
            path=floodwall.polygon_file,
        )

        # HydroMT function: get geodataframe from filename
        gdf_floodwall = self._model.data_catalog.get_geodataframe(
            polygon_file, geom=self._model.region, crs=self._model.crs
        )

        # Add floodwall attributes to geodataframe
        gdf_floodwall["name"] = floodwall.name
        if (gdf_floodwall.geometry.type == "MultiLineString").any():
            gdf_floodwall = gdf_floodwall.explode()

        try:
            heights = [
                float(
                    us.UnitfulLength(
                        value=float(height),
                        units=self.database.site.gui.units.default_length_units,
                    ).convert(us.UnitTypesLength("meters"))
                )
                for height in gdf_floodwall["z"]
            ]
            gdf_floodwall["z"] = heights
            self.logger.info("Using floodwall height from shape file.")
        except Exception:
            self.logger.warning(
                f"Could not use height data from file due to missing `z` column or missing values therein. Using uniform height of {floodwall.elevation} instead."
            )
            gdf_floodwall["z"] = floodwall.elevation.convert(
                us.UnitTypesLength(us.UnitTypesLength.meters)
            )

        # par1 is the overflow coefficient for weirs
        gdf_floodwall["par1"] = 0.6

        # HydroMT function: create floodwall
        self._model.setup_structures(structures=gdf_floodwall, stype="weir", merge=True)

    def _add_measure_greeninfra(self, green_infrastructure: GreenInfrastructure):
        # HydroMT function: get geodataframe from filename
        if green_infrastructure.selection_type == "polygon":
            polygon_file = resolve_filepath(
                ObjectDir.measure,
                green_infrastructure.name,
                green_infrastructure.polygon_file,
            )
        elif green_infrastructure.selection_type == "aggregation_area":
            # TODO this logic already exists in the Database controller but cannot be used due to cyclic imports
            # Loop through available aggregation area types
            for aggr_dict in self.database.site.fiat.config.aggregation:
                # check which one is used in measure
                if not aggr_dict.name == green_infrastructure.aggregation_area_type:
                    continue
                # load geodataframe
                aggr_areas = gpd.read_file(
                    db_path(TopLevelDir.static) / aggr_dict.file,
                    engine="pyogrio",
                ).to_crs(4326)
                # keep only aggregation area chosen
                polygon_file = aggr_areas.loc[
                    aggr_areas[aggr_dict.field_name]
                    == green_infrastructure.aggregation_area_name,
                    ["geometry"],
                ].reset_index(drop=True)
        else:
            raise ValueError(
                f"The selection type: {green_infrastructure.selection_type} is not valid"
            )

        gdf_green_infra = self._model.data_catalog.get_geodataframe(
            polygon_file,
            geom=self._model.region,
            crs=self._model.crs,
        )

        # Make sure no multipolygons are there
        gdf_green_infra = gdf_green_infra.explode()

        # HydroMT function: create storage volume
        self._model.setup_storage_volume(
            storage_locs=gdf_green_infra,
            volume=green_infrastructure.volume.convert(us.UnitTypesVolume.m3),
            merge=True,
        )

    def _add_measure_pump(self, pump: Pump):
        """Add pump to sfincs model.

        Parameters
        ----------
        pump : Pump
            pump information
        """
        polygon_file = resolve_filepath(ObjectDir.measure, pump.name, pump.polygon_file)
        # HydroMT function: get geodataframe from filename
        gdf_pump = self._model.data_catalog.get_geodataframe(
            polygon_file, geom=self._model.region, crs=self._model.crs
        )

        # HydroMT function: create floodwall
        self._model.setup_drainage_structures(
            structures=gdf_pump,
            stype="pump",
            discharge=pump.discharge.convert(us.UnitTypesDischarge.cms),
            merge=True,
        )

    ### SFINCS SETTERS ###
    def _set_single_river_forcing(self, discharge: IDischarge):
        """Add discharge to overland sfincs model.

        Parameters
        ----------
        discharge : IDischarge
            Discharge object with discharge timeseries data and river information.
        """
        if not isinstance(
            discharge, (DischargeConstant, DischargeSynthetic, DischargeCSV)
        ):
            self.logger.warning(
                f"Unsupported discharge forcing type: {discharge.__class__.__name__}"
            )
            return

        self.logger.info(f"Setting discharge forcing for river: {discharge.river.name}")

        time_frame = self.get_model_time()
        model_rivers = self._read_river_locations()

        # Check that the river is defined in the model and that the coordinates match
        river_loc = shapely.Point(
            discharge.river.x_coordinate, discharge.river.y_coordinate
        )
        tolerance = 0.001  # in degrees, ~111 meters at the equator. (0.0001: 11 meters at the equator)
        river_gdf = model_rivers[model_rivers.distance(river_loc) <= tolerance]
        river_inds = river_gdf.index.to_list()
        if len(river_inds) != 1:
            raise ValueError(
                f"River {discharge.river.name} is not defined in the sfincs model. Please ensure the river coordinates in the site.toml match the coordinates for rivers in the SFINCS model."
            )

        # Create a geodataframe with the river coordinates, the timeseries data and rename the column to the river index defined in the model
        if isinstance(discharge, DischargeCSV):
            df = discharge.to_dataframe(time_frame)
            conversion = us.UnitfulDischarge(value=1.0, units=discharge.units).convert(
                us.UnitTypesDischarge.cms
            )
        elif isinstance(discharge, DischargeConstant):
            df = discharge.to_dataframe(time_frame)
            conversion = us.UnitfulDischarge(
                value=1.0, units=discharge.discharge.units
            ).convert(us.UnitTypesDischarge.cms)
        elif isinstance(discharge, DischargeSynthetic):
            df = discharge.to_dataframe(time_frame)
            conversion = us.UnitfulDischarge(
                value=1.0, units=discharge.timeseries.peak_value.units
            ).convert(us.UnitTypesDischarge.cms)
        else:
            raise ValueError(
                f"Unsupported discharge forcing type: {discharge.__class__}"
            )

        df *= conversion

        df = df.rename(columns={df.columns[0]: river_inds[0]})

        # HydroMT function: set discharge forcing from time series and river coordinates
        self._model.setup_discharge_forcing(
            locations=river_gdf,
            timeseries=df,
            merge=True,
        )

    def _turn_off_bnd_press_correction(self):
        """Turn off the boundary pressure correction in the sfincs model."""
        self.logger.info(
            "Turning off boundary pressure correction in the offshore model"
        )
        self._model.set_config("pavbnd", -9999)

    def _set_waterlevel_forcing(self, df_ts: pd.DataFrame):
        """
        Add water level forcing to sfincs model.

        Values in the timeseries are expected to be relative to the main reference datum: `self.settings.water_level.reference`.
        The overland model reference: `self.settings.config.overland_model.reference` is used to convert the water levels to the reference of the overland model.

        Parameters
        ----------
        df_ts : pd.DataFrame
            Time series of water levels with the first column as the time index.


        """
        # Determine bnd points from reference overland model
        gdf_locs = self._read_waterlevel_boundary_locations()

        if len(df_ts.columns) == 1:
            # Go from 1 timeseries to timeseries for all boundary points
            name = df_ts.columns[0]
            for i in range(1, len(gdf_locs)):
                df_ts[i + 1] = df_ts[name]
            df_ts.columns = list(range(1, len(gdf_locs) + 1))

        # Datum
        sfincs_overland_reference_height = self.settings.water_level.get_datum(
            self.settings.config.overland_model.reference
        ).height.convert(us.UnitTypesLength.meters)

        df_ts -= sfincs_overland_reference_height

        # HydroMT function: set waterlevel forcing from time series
        self._model.set_forcing_1d(
            name="bzs", df_ts=df_ts, gdf_locs=gdf_locs, merge=False
        )

    # OFFSHORE
    def _add_pressure_forcing_from_grid(self, ds: xr.DataArray):
        """Add spatially varying barometric pressure to sfincs model.

        Parameters
        ----------
        ds : xr.DataArray
            - Required variables: ['press_msl' (Pa)]
            - Required coordinates: ['time', 'y', 'x']
            - spatial_ref: CRS
        """
        self.logger.info("Adding pressure forcing to the offshore model")
        self._model.setup_pressure_forcing_from_grid(press=ds)

    def _add_bzs_from_bca(self, event: Event, physical_projection: PhysicalProjection):
        # ONLY offshore models
        """Convert tidal constituents from bca file to waterlevel timeseries that can be read in by hydromt_sfincs."""
        if self.settings.config.offshore_model is None:
            raise ValueError("No offshore model found in sfincs config.")

        self.logger.info("Adding water level forcing to the offshore model")
        sb = SfincsBoundary()
        sb.read_flow_boundary_points(self.get_model_root() / "sfincs.bnd")
        sb.read_astro_boundary_conditions(self.get_model_root() / "sfincs.bca")

        times = pd.date_range(
            start=event.time.start_time,
            end=event.time.end_time,
            freq="10T",
        )

        # Predict tidal signal and add SLR
        if not sb.flow_boundary_points:
            raise ValueError("No flow boundary points found.")

        if self.settings.config.offshore_model.vertical_offset:
            correction = self.settings.config.offshore_model.vertical_offset.convert(
                us.UnitTypesLength.meters
            )
        else:
            correction = 0.0

        for bnd_ii in range(len(sb.flow_boundary_points)):
            tide_ii = (
                predict(sb.flow_boundary_points[bnd_ii].astro, times)
                + correction
                + physical_projection.sea_level_rise.convert(us.UnitTypesLength.meters)
            )

            if bnd_ii == 0:
                wl_df = pd.DataFrame(data={1: tide_ii}, index=times)
            else:
                wl_df[bnd_ii + 1] = tide_ii

        # Determine bnd points from reference overland model
        gdf_locs = self._read_waterlevel_boundary_locations()

        # HydroMT function: set waterlevel forcing from time series
        self._model.set_forcing_1d(
            name="bzs", df_ts=wl_df, gdf_locs=gdf_locs, merge=False
        )

    ### PRIVATE GETTERS ###
    def _get_result_path(self, scenario: Scenario) -> Path:
        """Return the path to store the results."""
        return self.database.scenarios.output_path / scenario.name / "Flooding"

    def _get_simulation_path(
        self, scenario: Scenario, sub_event: Optional[Event] = None
    ) -> Path:
        """
        Return the path to the simulation results.

        Parameters
        ----------
        scenario : Scenario
            The scenario for which to get the simulation path.
        sub_event : Optional[Event], optional
            The sub-event for which to get the simulation path, by default None.
            Is only used when the event associated with the scenario is an EventSet.
        """
        base_path = (
            self._get_result_path(scenario)
            / "simulations"
            / self.settings.config.overland_model.name
        )
        event = self.database.events.get(scenario.event)

        if isinstance(event, EventSet):
            if sub_event is None:
                raise ValueError("Event must be provided when scenario is an EventSet.")
            return base_path.parent / sub_event.name / base_path.name
        elif isinstance(event, Event):
            return base_path
        else:
            raise ValueError(f"Unsupported mode: {event.mode}")

    def _get_simulation_path_offshore(
        self, scenario: Scenario, sub_event: Optional[Event] = None
    ) -> Path:
        # Get the path to the offshore model (will not be used if offshore model is not created)
        if self.settings.config.offshore_model is None:
            raise ValueError("No offshore model found in sfincs config.")
        base_path = (
            self._get_result_path(scenario)
            / "simulations"
            / self.settings.config.offshore_model.name
        )
        event = self.database.events.get(scenario.event)
        if isinstance(event, EventSet):
            return base_path.parent / sub_event.name / base_path.name
        elif isinstance(event, Event):
            return base_path
        else:
            raise ValueError(f"Unsupported mode: {event.mode}")

    def _get_flood_map_paths(self, scenario: Scenario) -> list[Path]:
        """Return the paths to the flood maps that running this scenario should produce."""
        results_path = self._get_result_path(scenario)
        event = self.database.events.get(scenario.event)

        if isinstance(event, EventSet):
            map_fn = []
            for rp in self.database.site.fiat.risk.return_periods:
                map_fn.append(results_path / f"RP_{rp:04d}_maps.nc")
        elif isinstance(event, Event):
            map_fn = [results_path / "max_water_level_map.nc"]
        else:
            raise ValueError(f"Unsupported mode: {event.mode}")

        return map_fn

    def _get_event_input_path(self, event: Event) -> Path:
        """Return the path to the event input directory."""
        return self.database.events.input_path / event.name

    def _get_zsmax(self):
        """Read zsmax file and return absolute maximum water level over entire simulation."""
        self._model.read_results()
        zsmax = self._model.results["zsmax"].max(dim="timemax")
        zsmax.attrs["units"] = "m"
        return zsmax

    def _get_zs_points(self):
        """Read water level (zs) timeseries at observation points.

        Names are allocated from the site.toml.
        See also add_obs_points() above.
        """
        self._model.read_results()
        da = self._model.results["point_zs"]
        df = pd.DataFrame(index=pd.DatetimeIndex(da.time), data=da.to_numpy())

        names = []
        descriptions = []
        # get station names from site.toml
        if self.settings.obs_point is not None:
            obs_points = self.settings.obs_point
            for pt in obs_points:
                names.append(pt.name)
                descriptions.append(pt.description)

        pt_df = pd.DataFrame({"Name": names, "Description": descriptions})
        gdf = gpd.GeoDataFrame(
            pt_df,
            geometry=gpd.points_from_xy(da.point_x.values, da.point_y.values),
            crs=self._model.crs,
        )
        return df, gdf

    def _create_spw_file_from_track(
        self,
        track_forcing: Union[RainfallTrack, WindTrack],
        hurricane_translation: TranslationModel,
        name: str,
        output_dir: Path,
        include_rainfall: bool = False,
        recreate: bool = False,
    ):
        """
        Create a spiderweb file from a given TropicalCyclone track and save it to the event's input directory.

        Providing the output_dir argument allows to save the spiderweb file in a different directory.

        Parameters
        ----------
        output_dir : Path
            The directory where the spiderweb file is saved (or copied to if it already exists and recreate is False)
        recreate : bool, optional
            If True, the spiderweb file is recreated even if it already exists, by default False

        Returns
        -------
        Path
            the path to the created spiderweb file
        """
        if track_forcing.path is None:
            raise ValueError("No path to track file provided.")

        # Check file format
        match track_forcing.path.suffix:
            case ".spw":
                if recreate:
                    raise ValueError(
                        "Recreating spiderweb files from existing spiderweb files is not supported. Provide a track file instead."
                    )

                if track_forcing.path.exists():
                    return track_forcing.path

                elif (output_dir / track_forcing.path.name).exists():
                    return output_dir / track_forcing.path.name

                else:
                    raise FileNotFoundError(f"SPW file not found: {track_forcing.path}")
            case ".cyc":
                pass
            case _:
                raise ValueError(
                    "Track files should be in the DDB_CYC file format and must have .cyc extension, or in the SPW file format and must have .spw extension"
                )

        # Check if the spiderweb file already exists
        spw_file = output_dir / track_forcing.path.with_suffix(".spw").name
        if spw_file.exists():
            if recreate:
                os.remove(spw_file)
            else:
                return spw_file

        # Initialize the tropical cyclone
        tc = TropicalCyclone()
        tc.read_track(filename=str(track_forcing.path), fmt="ddb_cyc")

        # Alter the track of the tc if necessary
        tc = self._translate_tc_track(
            tc=tc, hurricane_translation=hurricane_translation
        )

        # Rainfall
        start = "Including" if include_rainfall else "Excluding"
        self.logger.info(f"{start} rainfall in the spiderweb file")
        tc.include_rainfall = include_rainfall

        self.logger.info(
            f"Creating spiderweb file for hurricane event `{name}`. This may take a while."
        )

        # Create spiderweb file from the track
        tc.to_spiderweb(spw_file)

        return spw_file

    def _translate_tc_track(
        self, tc: TropicalCyclone, hurricane_translation: TranslationModel
    ):
        if math.isclose(
            hurricane_translation.eastwest_translation.value, 0, abs_tol=1e-6
        ) and math.isclose(
            hurricane_translation.northsouth_translation.value, 0, abs_tol=1e-6
        ):
            return tc

        self.logger.info(f"Translating the track of the tropical cyclone `{tc.name}`")
        # First convert geodataframe to the local coordinate system
        crs = pyproj.CRS.from_string(self.settings.config.csname)
        tc.track = tc.track.to_crs(crs)

        # Translate the track in the local coordinate system
        tc.track["geometry"] = tc.track["geometry"].apply(
            lambda geom: translate(
                geom,
                xoff=hurricane_translation.eastwest_translation.convert(
                    us.UnitTypesLength.meters
                ),
                yoff=hurricane_translation.northsouth_translation.convert(
                    us.UnitTypesLength.meters
                ),
            )
        )

        # Convert the geodataframe to lat/lon
        tc.track = tc.track.to_crs(epsg=4326)

        return tc

    # @gundula do we keep this func, its not used anywhere?
    def _downscale_hmax(self, zsmax, demfile: Path):
        # read DEM and convert units to metric units used by SFINCS
        demfile_units = self.settings.dem.units
        dem_conversion = us.UnitfulLength(value=1.0, units=demfile_units).convert(
            us.UnitTypesLength("meters")
        )
        dem = dem_conversion * self._model.data_catalog.get_rasterdataset(demfile)
        dem = dem.rio.reproject(self._model.crs)

        # determine conversion factor for output floodmap
        floodmap_units = self.settings.config.floodmap_units
        floodmap_conversion = us.UnitfulLength(
            value=1.0, units=us.UnitTypesLength.meters
        ).convert(floodmap_units)

        hmax = utils.downscale_floodmap(
            zsmax=floodmap_conversion * zsmax,
            dep=floodmap_conversion * dem,
            hmin=0.01,
        )
        return hmax

    def _read_river_locations(self) -> gpd.GeoDataFrame:
        path = self.get_model_root() / "sfincs.src"

        with open(path) as f:
            lines = f.readlines()
        coords = [(float(line.split()[0]), float(line.split()[1])) for line in lines]
        points = [shapely.Point(coord) for coord in coords]

        return gpd.GeoDataFrame({"geometry": points}, crs=self._model.crs)

    def _read_waterlevel_boundary_locations(self) -> gpd.GeoDataFrame:
        with open(self.get_model_root() / "sfincs.bnd") as f:
            lines = f.readlines()
        coords = [(float(line.split()[0]), float(line.split()[1])) for line in lines]
        points = [shapely.Point(coord) for coord in coords]

        return gpd.GeoDataFrame({"geometry": points}, crs=self._model.crs)

    def _setup_sfincs_logger(self, model_root: Path) -> logging.Logger:
        """Initialize the logger for the SFINCS model."""
        # Create a logger for the SFINCS model manually
        sfincs_logger = logging.getLogger("SfincsModel")
        for handler in sfincs_logger.handlers[:]:
            sfincs_logger.removeHandler(handler)

        # Add a file handler
        file_handler = logging.FileHandler(
            filename=model_root.resolve() / "sfincs_model.log",
            mode="w",
        )
        sfincs_logger.setLevel(logging.DEBUG)
        sfincs_logger.addHandler(file_handler)
        return sfincs_logger

    def _cleanup_simulation_folder(
        self,
        path: Path,
        extensions: list[str] = [".spw"],
    ):
        """Remove all files with the given extensions in the given path."""
        if not path.exists():
            return

        for ext in extensions:
            for file in path.glob(f"*{ext}"):
                file.unlink()

    def _load_scenario_objects(self, scenario: Scenario, event: Event) -> None:
        self._scenario = scenario
        self._projection = self.database.projections.get(scenario.projection)
        self._strategy = self.database.strategies.get(scenario.strategy)
        self._event = event

        _event = self.database.events.get(scenario.event)
        if isinstance(_event, EventSet):
            self._event_set = _event
        else:
            self._event_set = None

    def _add_tide_gauge_plot(
        self, fig, event: Event, units: us.UnitTypesLength
    ) -> None:
        # check if event is historic
        if not isinstance(event, HistoricalEvent):
            return
        if self.settings.tide_gauge is None:
            return
        df_gauge = self.settings.tide_gauge.get_waterlevels_in_time_frame(
            time=TimeFrame(
                start_time=event.time.start_time,
                end_time=event.time.end_time,
            ),
            units=us.UnitTypesLength(units),
        )

        if df_gauge is not None:
            gauge_reference_height = self.settings.water_level.get_datum(
                self.settings.tide_gauge.reference
            ).height.convert(units)

            waterlevel = df_gauge.iloc[:, 0] + gauge_reference_height

            # If data is available, add to plot
            fig.add_trace(
                go.Scatter(
                    x=pd.DatetimeIndex(df_gauge.index),
                    y=waterlevel,
                    line_color="#ea6404",
                )
            )
            fig["data"][0]["name"] = "model"
            fig["data"][1]["name"] = "measurement"
            fig.update_layout(showlegend=True)



from abc import abstractmethod
from pathlib import Path

from flood_adapt.misc.database_user import DatabaseUser
from flood_adapt.objects.scenarios.scenarios import Scenario


class IAdapter(DatabaseUser):
    """Adapter interface for all models run in FloodAdapt."""

    @abstractmethod
    def has_run(self, scenario: Scenario) -> bool:
        """Return True if the model has been run."""
        pass

    @abstractmethod
    def __enter__(self) -> "IAdapter":
        """Use the adapter as a context manager to handle opening and closing of the model and attached resources like logfiles.

        Returns
        -------
            self: the adapter object

        Usage
        -----
        >>> with Adapter(...) as model:
        >>>     ...
        >>>     model.get_result(...)
        >>>     model.run(...)

        Entering the with block will call adapter.__enter__() and
        Exiting the with block (via regular execution or an error) will call adapter.__exit__()
        """
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        """Close the model and release any resources.

        Usage
        -----
        >>> with Adapter as model:
        >>>     ...
        >>>     model.run()

        Entering the `with` block will call adapter.__enter__()
        Exiting the `with` block (via regular execution or an error) will call adapter.__exit__()

        Returns
        -------
            False to propagate/reraise any exceptions that occurred in the with block
            True to suppress any exceptions that occurred in the with block
        """
        pass

    @abstractmethod
    def read(self, path: Path):
        """Read the model configuration from a path or other source."""
        pass

    @abstractmethod
    def write(self, path_out: Path, overwrite: bool = True):
        """Write the current model configuration to a path or other destination."""
        pass

    @abstractmethod
    def run(self, scenario: Scenario):
        """Perform the whole workflow (preprocess, execute and postprocess) of running the model."""
        pass

    @abstractmethod
    def preprocess(self, scenario: Scenario):
        """Prepare the model for execution."""
        pass

    @abstractmethod
    def execute(self, path: Path, strict: bool = True) -> bool:
        """Run the model kernel at the specified path.

        Returns True if the model ran successfully, False otherwise.

        If strict is True, raise an exception if the model fails to run.
        """
        pass

    @abstractmethod
    def postprocess(self, scenario: Scenario):
        """Process the model output."""
        pass



import os
from pathlib import Path

import tomli
import tomli_w
from pydantic import BaseModel, Field


class Object(BaseModel):
    """Base class for FloodAdapt objects.

    Attributes
    ----------
    name : str
        Name of the object.
    description : str
        Description of the object. defaults to "".
    """

    name: str = Field(
        ...,
        description="Name of the object.",
        min_length=1,
        pattern='^[^<>:"/\\\\|?* ]*$',
    )
    description: str = Field(default="", description="Description of the object.")

    @classmethod
    def load_file(cls, file_path: Path | str | os.PathLike) -> "Object":
        """Load object from file.

        Parameters
        ----------
        file_path : Path | str | os.PathLike
            Path to the file to load.

        """
        with open(file_path, mode="rb") as fp:
            toml = tomli.load(fp)
        return cls.model_validate(toml)

    def save(self, toml_path: Path | str | os.PathLike) -> None:
        """Save object to disk.

        Parameters
        ----------
        toml_path : Path | str | os.PathLike
            Path to the file to save.

        """
        self.save_additional(output_dir=Path(toml_path).parent)
        with open(toml_path, "wb") as f:
            tomli_w.dump(self.model_dump(exclude_none=True), f)

    def save_additional(self, output_dir: Path | str | os.PathLike) -> None:
        """Save additional files to database if the object has any and update attrs to reflect the change in file location.

        This method should be overridden if the object has additional files.
        """
        pass

    def __eq__(self, value):
        if not isinstance(value, self.__class__):
            # don't attempt to compare against unrelated types
            return False
        _self = self.model_dump(exclude={"name", "description"}, exclude_none=True)
        _other = value.model_dump(exclude={"name", "description"}, exclude_none=True)
        return _self == _other



from flood_adapt import __version__
from flood_adapt.dbs_classes.interface.database import IDatabase
from flood_adapt.misc.log import FloodAdaptLogging
from flood_adapt.misc.utils import finished_file_exists, write_finished_file
from flood_adapt.objects.scenarios.scenarios import Scenario
from flood_adapt.workflows.impacts_integrator import Impacts


class ScenarioRunner:
    """class holding all information related to a scenario."""

    def __init__(self, database: IDatabase, scenario: Scenario) -> None:
        """Create a Scenario object."""
        self.logger = FloodAdaptLogging.getLogger("ScenarioRunner")
        self._database = database
        self._load_objects(scenario)
        self.site_info = self._database.site
        self.results_path = self._database.scenarios.output_path / self._scenario.name

    def _load_objects(self, scenario: Scenario) -> None:
        """Load objects from the database."""
        self._scenario = scenario
        self._event = self._database.events.get(scenario.event)
        self._projection = self._database.projections.get(scenario.projection)
        self._strategy = self._database.strategies.get(scenario.strategy)

    @property
    def impacts(self) -> Impacts:
        return Impacts(
            scenario=self._scenario,
        )

    def run(self) -> None:
        """Run hazard and impact models for the scenario."""
        self._database.has_run_hazard(self._scenario.name)
        self._load_objects(self._scenario)
        self.results_path.mkdir(parents=True, exist_ok=True)

        # Initiate the logger for all the integrator scripts.
        log_file = self.results_path.joinpath(f"logfile_{self._scenario.name}.log")
        with FloodAdaptLogging.to_file(file_path=log_file):
            self.logger.info(f"FloodAdapt version `{__version__}`")
            self.logger.info(f"Started evaluation of `{self._scenario.name}`")

            hazard_models = [
                self._database.static.get_overland_sfincs_model(),
            ]
            for hazard in hazard_models:
                if not hazard.run_completed(self._scenario):
                    hazard.run(self._scenario)
                else:
                    self.logger.info(
                        f"Hazard for scenario '{self._scenario.name}' has already been run."
                    )

            if not self.impacts.has_run:
                self.impacts.run()
            else:
                self.logger.info(
                    f"Impacts for scenario `{self._scenario.name}` has already been run."
                )

            self.logger.info(f"Finished evaluation of `{self._scenario.name}`")

        # write finished file to indicate that the scenario has been run
        write_finished_file(self.results_path)

    def has_run_check(self):
        """Check if the scenario has been run."""
        return finished_file_exists(self.results_path)



import shutil
from typing import Any

from flood_adapt.dbs_classes.dbs_template import DbsTemplate
from flood_adapt.misc.utils import finished_file_exists
from flood_adapt.objects.scenarios.scenarios import Scenario
from flood_adapt.workflows.benefit_runner import BenefitRunner


class DbsScenario(DbsTemplate[Scenario]):
    dir_name = "scenarios"
    display_name = "Scenario"
    _object_class = Scenario

    def summarize_objects(self) -> dict[str, list[Any]]:
        """Return a dictionary with info on the events that currently exist in the database.

        Returns
        -------
        dict[str, Any]
            Includes 'name', 'description', 'path' and 'last_modification_date' info
        """
        scenarios = super().summarize_objects()
        scenarios["Projection"] = [
            self._read_variable_in_toml("projection", path)
            for path in scenarios["path"]
        ]
        scenarios["Event"] = [
            self._read_variable_in_toml("event", path) for path in scenarios["path"]
        ]
        scenarios["Strategy"] = [
            self._read_variable_in_toml("strategy", path) for path in scenarios["path"]
        ]
        scenarios["finished"] = [self.has_run_check(scn) for scn in scenarios["name"]]

        return scenarios

    def delete(self, name: str, toml_only: bool = False):
        """Delete an already existing scenario in the database.

        Parameters
        ----------
        name : str
            name of the scenario to be deleted
        toml_only : bool, optional
            whether to only delete the toml file or the entire folder. If the folder is empty after deleting the toml,
            it will always be deleted. By default False

        Raises
        ------
        ValueError
            Raise error if scenario to be deleted is already in use.
        """
        # First delete the scenario
        super().delete(name, toml_only)

        # Then delete the results
        if (self.output_path / name).exists():
            shutil.rmtree(self.output_path / name, ignore_errors=False)

    def check_higher_level_usage(self, name: str) -> list[str]:
        """Check if a scenario is used in a benefit.

        Parameters
        ----------
        name : str
            name of the scenario to be checked

        Returns
        -------
            list[str]
                list of benefits that use the scenario
        """
        benefits = [
            self._database.benefits.get(benefit)
            for benefit in self._database.benefits.summarize_objects()["name"]
        ]
        used_in_benefit = []
        for benefit in benefits:
            runner = BenefitRunner(database=self._database, benefit=benefit)
            scenarios = runner.check_scenarios()["scenario created"].to_list()
            for scenario in scenarios:
                if name == scenario:
                    used_in_benefit.append(benefit.name)

        return used_in_benefit

    def equal_hazard_components(self, left: Scenario, right: Scenario) -> bool:
        """Check if two scenarios have the same hazard components.

        Parameters
        ----------
        left : Scenario
            first scenario to be compared
        right : Scenario
            second scenario to be compared

        Returns
        -------
            bool
                True if the scenarios have the same hazard components, False otherwise
        """
        event_left = self._database.events.get(left.event)
        event_right = self._database.events.get(right.event)
        equal_events = event_left == event_right

        left_projection = self._database.projections.get(left.projection)
        right_projection = self._database.projections.get(right.projection)
        equal_projection = (
            left_projection.physical_projection == right_projection.physical_projection
        )

        left_strategy = self._database.strategies.get(
            left.strategy
        ).get_hazard_strategy()
        right_strategy = self._database.strategies.get(
            right.strategy
        ).get_hazard_strategy()
        equal_strategy = left_strategy == right_strategy

        return equal_events and equal_projection and equal_strategy

    def has_run_check(self, name: str) -> bool:
        """Check if the scenario has been run.

        Parameters
        ----------
        name : str
            name of the scenario to be checked

        Returns
        -------
            bool
                True if the scenario has been run, False otherwise
        """
        results_path = self.output_path / name
        return finished_file_exists(results_path)



import os
from pathlib import Path

from flood_adapt.config.sfincs import FloodmapType
from flood_adapt.misc.database_user import DatabaseUser
from flood_adapt.misc.log import FloodAdaptLogging
from flood_adapt.objects.events.event_set import EventSet
from flood_adapt.objects.events.events import Mode
from flood_adapt.objects.strategies.strategies import Strategy


class FloodMap(DatabaseUser):
    logger = FloodAdaptLogging.getLogger("FloodMap")

    type: FloodmapType

    name: str
    path: Path | os.PathLike | list[Path | os.PathLike]
    event_set: EventSet

    def __init__(self, scenario_name: str) -> None:
        self.name = scenario_name
        self.type = self.database.site.fiat.config.floodmap_type
        self._get_flood_map_paths()

    def _get_flood_map_paths(self):
        base_dir = self.database.scenarios.output_path / self.name / "Flooding"
        # TODO check naming of files
        if self.mode == Mode.single_event:
            if self.type == FloodmapType.water_level:
                self.path = base_dir / "max_water_level_map.nc"
            elif self.type == FloodmapType.water_depth:
                self.path = base_dir / f"FloodMap_{self.name}.tif"
        elif self.mode == Mode.risk:
            if self.type == FloodmapType.water_level:
                self.path = list(base_dir.glob("RP_*_maps.nc"))
            elif self.type == FloodmapType.water_depth:
                self.path = list(base_dir.glob("RP_*_maps.tif"))

    @property
    def has_run(self) -> bool:
        if self.mode == Mode.single_event:
            return self.path.exists()
        elif self.mode == Mode.risk:
            check_files = [RP_map.exists() for RP_map in self.path]
            check_rps = len(self.path) == len(
                self.database.site.fiat.risk.return_periods
            )
            return all(check_files) & check_rps

    @property
    def scenario(self):
        if hasattr(self, "_scenario"):
            return self._scenario
        self._scenario = self.database.scenarios.get(self.name)
        return self._scenario

    @property
    def mode(self):
        if hasattr(self, "_mode"):
            return self._mode
        self._mode = self.database.events.get(self.scenario.event).mode
        return self._mode

    @property
    def crs(self):
        if hasattr(self, "_crs"):
            return self._crs
        self._crs = self.database.site.crs
        return self._crs

    @property
    def hazard_strategy(self) -> Strategy:
        return self.database.strategies.get(
            self.scenario.strategy
        ).get_hazard_strategy()

    @property
    def physical_projection(self):
        if hasattr(self, "_physical_projection"):
            return self._physical_projection
        self._physical_projection = self.database.projections.get(
            self.scenario.projection
        ).physical_projection
        return self._physical_projection



import shutil
from pathlib import Path

import pandas as pd

from flood_adapt.adapter.interface.offshore import IOffshoreSfincsHandler
from flood_adapt.adapter.sfincs_adapter import SfincsAdapter
from flood_adapt.misc.database_user import DatabaseUser
from flood_adapt.misc.log import FloodAdaptLogging
from flood_adapt.misc.path_builder import (
    ObjectDir,
    TopLevelDir,
    db_path,
)
from flood_adapt.objects.events.event_set import EventSet
from flood_adapt.objects.events.events import Event, Mode
from flood_adapt.objects.events.historical import HistoricalEvent
from flood_adapt.objects.forcing.forcing import (
    ForcingSource,
    IWind,
)
from flood_adapt.objects.forcing.meteo_handler import MeteoHandler
from flood_adapt.objects.forcing.wind import WindMeteo
from flood_adapt.objects.scenarios.scenarios import Scenario


class OffshoreSfincsHandler(IOffshoreSfincsHandler, DatabaseUser):
    logger = FloodAdaptLogging.getLogger("OffshoreSfincsAdapter")
    template_path: Path

    def __init__(self, scenario: Scenario, event: Event) -> None:
        self.template_path = (
            self.database.static.get_offshore_sfincs_model().get_model_root()
        )
        self.scenario = scenario
        if isinstance(event, EventSet):
            raise ValueError(
                "OffshoreSfincsHandler does not support EventSets. Provide the sub events directly "
            )
        self.event = event

    def get_resulting_waterlevels(self) -> pd.DataFrame:
        """Get the water levels from the offshore model.

        Note that the returned water levels are relative to the reference datum of the offshore model.
        To convert to a different datum, add the offshore reference datum height and subtract the desired reference datum height.

        Returns
        -------
        pd.DataFrame
            A DataFrame with the water levels for each boundary condition point. Relative to the reference datum of the offshore model.

        """
        path = self._get_simulation_path()
        if not self.requires_offshore_run(self.event):
            raise ValueError("Offshore model is not required for this event")

        self.run_offshore()

        with SfincsAdapter(model_root=path) as offshore_model:
            waterlevels = offshore_model.get_wl_df_from_offshore_his_results()

        return waterlevels

    @staticmethod
    def requires_offshore_run(event: Event) -> bool:
        return any(
            forcing.source in [ForcingSource.MODEL, ForcingSource.TRACK]
            for forcing in event.get_forcings()
        )

    def run_offshore(self):
        """Prepare the forcings of the historical event.

        If the forcings require it, this function will:
        - preprocess and run offshore model: prepare and run the offshore model to obtain water levels for the boundary condition of the nearshore model.

        """
        sim_path = self._get_simulation_path()

        sim_path.mkdir(parents=True, exist_ok=True)
        self._preprocess_sfincs_offshore()
        self._execute_sfincs_offshore(sim_path)

    def _preprocess_sfincs_offshore(self):
        """Preprocess offshore model to obtain water levels for boundary condition of the nearshore model.

        This function is reused for ForcingSources: MODEL & TRACK.

        Args:
            sim_path path to the root of the offshore model
        """
        self.logger.info(
            f"Preparing offshore model to generate waterlevels for `{self.scenario.name}`"
        )
        sim_path = self._get_simulation_path()
        # SfincsAdapter.write() doesnt write the bca file apparently so we need to copy the template
        if sim_path.exists():
            shutil.rmtree(sim_path)
        shutil.copytree(self.template_path, sim_path)

        with SfincsAdapter(model_root=sim_path) as _offshore_model:
            if _offshore_model.sfincs_completed(sim_path):
                _offshore_model.logger.info(
                    f"Skip preprocessing offshore model as it has already been run for `{self.scenario.name}`."
                )
                return
            # Load objects, set root & write template model
            _offshore_model._load_scenario_objects(self.scenario, self.event)
            _offshore_model.write(path_out=sim_path)
            _offshore_model.set_timing(self.event.time)

            # Add water levels
            _offshore_model._add_bzs_from_bca(
                _offshore_model._event, _offshore_model._projection.physical_projection
            )

            # Add spw if applicable
            track_forcings = [
                f
                for f in _offshore_model._event.get_forcings()
                if f.source == ForcingSource.TRACK
            ]
            if track_forcings:
                for forcing in track_forcings:
                    _offshore_model.add_forcing(forcing)

            # Add wind and if applicable pressure forcing from meteo data
            elif isinstance(_offshore_model._event, HistoricalEvent):
                wind_forcings = [
                    f
                    for f in _offshore_model._event.get_forcings()
                    if isinstance(f, IWind)
                ]

                if wind_forcings:
                    if len(wind_forcings) > 1:
                        raise ValueError("Only one wind forcing is allowed")
                    wind_forcing = wind_forcings[0]

                    # Add wind forcing
                    if wind_forcing not in track_forcings:
                        _offshore_model.add_forcing(wind_forcing)

                    # Add pressure forcing for the offshore model (this doesnt happen normally in _add_forcing_wind() for overland models)
                    if isinstance(wind_forcing, WindMeteo):
                        ds = MeteoHandler().read(_offshore_model._event.time)
                        _offshore_model._add_pressure_forcing_from_grid(ds=ds)

            # write sfincs model in output destination
            _offshore_model.write(path_out=sim_path)

    def _execute_sfincs_offshore(self, sim_path: Path):
        self.logger.info(f"Running offshore model in {sim_path}")
        sim_path = self._get_simulation_path()
        with SfincsAdapter(model_root=sim_path) as _offshore_model:
            if _offshore_model.sfincs_completed(sim_path):
                self.logger.info(
                    "Skip running offshore model as it has already been run."
                )
                return
            try:
                _offshore_model.execute(path=sim_path)
            except RuntimeError as e:
                raise RuntimeError(
                    f"Failed to run offshore model for {self.scenario.name}"
                ) from e

    def _get_simulation_path(self) -> Path:
        main_event = self.database.events.get(self.scenario.event)
        if main_event.mode == Mode.risk:
            return (
                db_path(
                    TopLevelDir.output,
                    object_dir=ObjectDir.scenario,
                    obj_name=self.scenario.name,
                )
                / "Flooding"
                / "simulations"
                / self.event.name
                / self.template_path.name
            )
        else:
            return (
                db_path(
                    TopLevelDir.output,
                    object_dir=ObjectDir.scenario,
                    obj_name=self.scenario.name,
                )
                / "Flooding"
                / "simulations"
                / self.template_path.name
            )



from pathlib import Path
from typing import Any, Callable, Tuple, Union

import geopandas as gpd
import pandas as pd
from cht_cyclones.cyclone_track_database import CycloneTrackDatabase

from flood_adapt.adapter.fiat_adapter import FiatAdapter
from flood_adapt.adapter.sfincs_adapter import SfincsAdapter
from flood_adapt.config.config import Settings
from flood_adapt.dbs_classes.interface.database import IDatabase
from flood_adapt.dbs_classes.interface.static import IDbsStatic


def cache_method_wrapper(func: Callable) -> Callable:
    def wrapper(self, *args: Tuple[Any], **kwargs: dict[str, Any]) -> Any:
        if func.__name__ not in self._cached_data:
            self._cached_data[func.__name__] = {}

        args_key = (
            str(args) + str(sorted(kwargs.items())) if args or kwargs else "no_args"
        )
        if args_key in self._cached_data[func.__name__]:
            return self._cached_data[func.__name__][args_key]

        result = func(self, *args, **kwargs)
        self._cached_data[func.__name__][args_key] = result

        return result

    return wrapper


class DbsStatic(IDbsStatic):
    _cached_data: dict[str, Any] = {}
    _database: IDatabase

    def __init__(self, database: IDatabase):
        """Initialize any necessary attributes."""
        self._database = database

    def load_static_data(self):
        """Read data into the cache.

        This is used to read data from the database and store it in the cache.
        """
        self.get_aggregation_areas()
        self.get_model_boundary()
        self.get_model_grid()
        self.get_obs_points()
        self.get_slr_scn_names()
        self.get_buildings()
        self.get_property_types()

    @cache_method_wrapper
    def get_aggregation_areas(self) -> dict[str, gpd.GeoDataFrame]:
        """Get a list of the aggregation areas that are provided in the site configuration.

        These are expected to much the ones in the FIAT model.

        Returns
        -------
        list[gpd.GeoDataFrame]
            list of gpd.GeoDataFrames with the polygons defining the aggregation areas
        """
        aggregation_areas = {}
        for aggr_dict in self._database.site.fiat.config.aggregation:
            aggregation_areas[aggr_dict.name] = gpd.read_file(
                self._database.static_path / aggr_dict.file,
                engine="pyogrio",
            ).to_crs(4326)
            # Use always the same column name for name labels
            aggregation_areas[aggr_dict.name] = aggregation_areas[
                aggr_dict.name
            ].rename(columns={aggr_dict.field_name: "name"})
            # Make sure they are ordered alphabetically
            aggregation_areas[aggr_dict.name].sort_values(by="name").reset_index(
                drop=True
            )
        return aggregation_areas

    @cache_method_wrapper
    def get_model_boundary(self) -> gpd.GeoDataFrame:
        """Get the model boundary from the SFINCS model."""
        bnd = self.get_overland_sfincs_model().get_model_boundary()
        return bnd

    @cache_method_wrapper
    def get_model_grid(self):
        """Get the model grid from the SFINCS model.

        Returns
        -------
        QuadtreeGrid
            The model grid
        """
        grid = self.get_overland_sfincs_model().get_model_grid()
        return grid

    @cache_method_wrapper
    def get_obs_points(self) -> gpd.GeoDataFrame:
        """Get the observation points from the flood hazard model."""
        names = []
        descriptions = []
        lat = []
        lon = []
        if self._database.site.sfincs.obs_point is not None:
            obs_points = self._database.site.sfincs.obs_point
            for pt in obs_points:
                names.append(pt.name)
                descriptions.append(pt.description)
                lat.append(pt.lat)
                lon.append(pt.lon)

        # create gpd.GeoDataFrame from obs_points in site file
        df = pd.DataFrame({"name": names, "description": descriptions})
        # TODO: make crs flexible and add this as a parameter to site.toml?
        gdf = gpd.GeoDataFrame(
            df, geometry=gpd.points_from_xy(lon, lat), crs="EPSG:4326"
        )
        return gdf

    @cache_method_wrapper
    def get_static_map(self, path: Union[str, Path]) -> gpd.GeoDataFrame:
        """Get a map from the static folder.

        Parameters
        ----------
        path : Union[str, Path]
            Path to the map relative to the static folder

        Returns
        -------
        gpd.GeoDataFrame
            gpd.GeoDataFrame with the map in crs 4326

        Raises
        ------
        FileNotFoundError
            If the file is not found
        """
        # Read the map
        full_path = self._database.static_path / path
        if not full_path.is_file():
            raise FileNotFoundError(f"File {full_path} not found")
        return gpd.read_file(full_path, engine="pyogrio").to_crs(4326)

    @cache_method_wrapper
    def get_slr_scn_names(self) -> list:
        """Get the names of the sea level rise scenarios from the file provided.

        Returns
        -------
        list
            List of scenario names
        """
        input_file = self._database.static_path.joinpath(
            self._database.site.sfincs.slr_scenarios.file
        )
        df = pd.read_csv(input_file)
        names = df.columns[2:].to_list()
        return names

    @cache_method_wrapper
    def get_green_infra_table(self, measure_type: str) -> pd.DataFrame:
        """Return a table with different types of green infrastructure measures and their infiltration depths.

        This is read by a csv file in the database.

        Returns
        -------
        pd.DataFrame
            Table with values
        """
        # Read file from database
        df = pd.read_csv(
            self._database.static_path.joinpath(
                "green_infra_table", "green_infra_lookup_table.csv"
            )
        )

        # Get column with values
        val_name = "Infiltration depth"
        col_name = [name for name in df.columns if val_name in name][0]
        if not col_name:
            raise KeyError(f"A column with a name containing {val_name} was not found!")

        # Get list of types per measure
        df["types"] = [
            [x.strip() for x in row["types"].split(",")] for i, row in df.iterrows()
        ]

        # Show specific values based on measure type
        inds = [i for i, row in df.iterrows() if measure_type in row["types"]]
        df = df.drop(columns="types").iloc[inds, :]

        return df

    @cache_method_wrapper
    def get_buildings(self) -> gpd.GeoDataFrame:
        """Get the building footprints from the FIAT model.

        This should only be the buildings excluding any other types (e.g., roads)
        The parameters non_building_names in the site config is used for that.

        Returns
        -------
        gpd.GeoDataFrame
            building footprints with all the FIAT columns
        """
        return self.get_fiat_model().get_buildings()

    @cache_method_wrapper
    def get_property_types(self) -> list:
        """_summary_.

        Returns
        -------
        list
            _description_
        """
        return self.get_fiat_model().get_property_types()

    def get_overland_sfincs_model(self) -> SfincsAdapter:
        """Get the template offshore SFINCS model."""
        overland_path = (
            self._database.static_path
            / "templates"
            / self._database.site.sfincs.config.overland_model.name
        )
        with SfincsAdapter(model_root=overland_path) as overland_model:
            return overland_model

    def get_offshore_sfincs_model(self) -> SfincsAdapter:
        """Get the template overland Sfincs model."""
        if self._database.site.sfincs.config.offshore_model is None:
            raise ValueError("No offshore model defined in the site configuration.")

        offshore_path = (
            self._database.static_path
            / "templates"
            / self._database.site.sfincs.config.offshore_model.name
        )
        with SfincsAdapter(model_root=offshore_path) as offshore_model:
            return offshore_model

    def get_fiat_model(self) -> FiatAdapter:
        """Get the path to the FIAT model."""
        if self._database.site.fiat is None:
            raise ValueError("No FIAT model defined in the site configuration.")
        template_path = self._database.static_path / "templates" / "fiat"
        with FiatAdapter(
            model_root=template_path,
            config=self._database.site.fiat.config,
            exe_path=Settings().fiat_path,
            delete_crashed_runs=Settings().delete_crashed_runs,
            config_base_path=self._database.static_path,
        ) as fm:
            return fm

    @cache_method_wrapper
    def get_cyclone_track_database(self) -> CycloneTrackDatabase:
        if self._database.site.sfincs.cyclone_track_database is None:
            raise ValueError(
                "No cyclone track database defined in the site configuration."
            )
        database_file = str(
            self._database.static_path
            / "cyclone_track_database"
            / self._database.site.sfincs.cyclone_track_database.file
        )
        return CycloneTrackDatabase("ibtracs", file_name=database_file)



from pathlib import Path
from typing import Any, List

import tomli

from flood_adapt.objects.events.event_set import EventSet
from flood_adapt.objects.events.events import (
    Event,
    Mode,
    Template,
)
from flood_adapt.objects.events.historical import HistoricalEvent
from flood_adapt.objects.events.hurricane import (
    HurricaneEvent,
    TranslationModel,
)
from flood_adapt.objects.events.synthetic import SyntheticEvent

__all__ = ["TranslationModel", "EventSet"]


class EventFactory:
    """Factory class for creating events.

    This class is used to create events based on a template.

    Attributes
    ----------
    _EVENT_TEMPLATES : dict[str, (Event, Event)]
        Dictionary mapping event templates to event classes and models
    """

    _EVENT_TEMPLATES = {
        Template.Hurricane: HurricaneEvent,
        Template.Historical: HistoricalEvent,
        Template.Synthetic: SyntheticEvent,
    }

    @staticmethod
    def get_event_from_template(template: Template) -> type[Event]:
        """Get the event class corresponding to the template.

        Parameters
        ----------
        template : str
            Name of the event template

        Returns
        -------
        Type[Event]
            Event template
        """
        if template not in EventFactory._EVENT_TEMPLATES:
            raise ValueError(f"Invalid event template: {template}")
        return EventFactory._EVENT_TEMPLATES[template]

    @staticmethod
    def read_template(filepath: Path) -> Template:
        """Get event template from toml file."""
        if not filepath.exists():
            raise FileNotFoundError(f"File not found: {filepath}")
        with open(filepath, mode="rb") as fp:
            toml = tomli.load(fp)
        if (template := toml.get("template")) is None:
            raise ValueError(f"Event template not found in {filepath}")

        return Template(template)

    @staticmethod
    def read_mode(filepath: Path) -> Mode:
        """Get event mode from toml file."""
        if not filepath.exists():
            raise FileNotFoundError(f"File not found: {filepath}")
        with open(filepath, mode="rb") as fp:
            toml = tomli.load(fp)
        if toml.get("mode") is None:
            raise ValueError(f"Event mode not found in {filepath}")
        return Mode(toml.get("mode"))

    @staticmethod
    def load_file(toml_file: Path) -> Event:
        """Return event object based on toml file.

        Parameters
        ----------
        toml_file : str
            Template name

        Returns
        -------
        Event
            Event object
        """
        mode = EventFactory.read_mode(toml_file)
        if mode == Mode.risk:
            event_type = EventSet
        elif mode == Mode.single_event:
            template = Template(EventFactory.read_template(toml_file))
            event_type = EventFactory.get_event_from_template(template)
        else:
            raise ValueError(f"Invalid event mode: {mode}")
        return event_type.load_file(toml_file)

    @staticmethod
    def load_dict(attrs: dict[str, Any] | Event) -> Event | EventSet:
        """Return event object based on attrs dict.

        Parameters
        ----------
        attrs : dict[str, Any]
            Event attributes

        Returns
        -------
        Event
            Event object based on template
        """
        if isinstance(attrs, Event):
            mode = attrs.mode
            template = attrs.template
        else:
            mode = Mode(attrs.get("mode"))
            template = Template(attrs.get("template"))

        if mode == Mode.risk:
            # TODO Load events
            return EventSet(**attrs)
        elif mode == Mode.single_event:
            return EventFactory.get_event_from_template(template)(**attrs)
        else:
            raise ValueError(f"Invalid event mode: {mode}")

    @staticmethod
    def get_allowed_forcings(template) -> dict[str, List[str]]:
        return EventFactory.get_event_from_template(template).get_allowed_forcings()



from pathlib import Path
from typing import Any

import geopandas as gpd

from flood_adapt.dbs_classes.dbs_template import DbsTemplate
from flood_adapt.misc.utils import resolve_filepath
from flood_adapt.objects.measures.measure_factory import MeasureFactory
from flood_adapt.objects.measures.measures import Measure


class DbsMeasure(DbsTemplate[Measure]):
    dir_name = "measures"
    display_name = "Measure"
    _object_class = Measure

    def get(self, name: str) -> Measure:
        """Return a measure object.

        Parameters
        ----------
        name : str
            name of the measure to be returned

        Returns
        -------
        Measure
            measure object
        """
        # Make the full path to the object
        full_path = self.input_path / name / f"{name}.toml"

        # Check if the object exists
        if not Path(full_path).is_file():
            raise ValueError(f"{self.display_name}: '{name}' does not exist.")

        # Load and return the object
        return MeasureFactory.get_measure_object(full_path)

    def summarize_objects(self) -> dict[str, list[Any]]:
        """Return a dictionary with info on the measures that currently exist in the database.

        Returns
        -------
        dict[str, Any]
            Includes 'name', 'description', 'path' and 'last_modification_date' and 'geometry' info
        """
        measures = self._get_object_summary()
        objects = [self.get(name) for name in measures["name"]]

        geometries = []
        for obj in objects:
            # If polygon is used read the polygon file
            if hasattr(obj, "polygon_file") and obj.polygon_file:
                src_path = resolve_filepath(
                    object_dir=self.dir_name,
                    obj_name=obj.name,
                    path=obj.polygon_file,
                )
                geometries.append(gpd.read_file(src_path))
            # If aggregation area is used read the polygon from the aggregation area name
            elif hasattr(obj, "aggregation_area_name") and obj.aggregation_area_name:
                if (
                    obj.aggregation_area_type
                    not in self._database.static.get_aggregation_areas()
                ):
                    raise ValueError(
                        f"Aggregation area type {obj.aggregation_area_type} for measure {obj.name} does not exist."
                    )
                gdf = self._database.static.get_aggregation_areas()[
                    obj.aggregation_area_type
                ]
                if obj.aggregation_area_name not in gdf["name"].to_numpy():
                    raise ValueError(
                        f"Aggregation area name {obj.aggregation_area_name} for measure {obj.name} does not exist."
                    )
                geometries.append(gdf.loc[gdf["name"] == obj.aggregation_area_name, :])
            # Else assign a None value
            else:
                geometries.append(None)

        measures["geometry"] = geometries
        return measures

    def check_higher_level_usage(self, name: str) -> list[str]:
        """Check if a measure is used in a strategy.

        Parameters
        ----------
        name : str
            name of the measure to be checked

        Returns
        -------
        list[str]
            list of strategies that use the measure
        """
        # Get all the strategies
        strategies = [
            self._database.strategies.get(strategy)
            for strategy in self._database.strategies.summarize_objects()["name"]
        ]

        # Check if measure is used in a strategy
        used_in_strategy = [
            strategy.name
            for strategy in strategies
            for measure in strategy.measures
            if name == measure
        ]
        return used_in_strategy



from pathlib import Path

from flood_adapt.misc.database_user import DatabaseUser
from flood_adapt.misc.log import FloodAdaptLogging
from flood_adapt.misc.path_builder import (
    ObjectDir,
    TopLevelDir,
    db_path,
)
from flood_adapt.objects.projections.projections import SocioEconomicChange
from flood_adapt.objects.scenarios.scenarios import Scenario
from flood_adapt.objects.strategies.strategies import Strategy
from flood_adapt.workflows.floodmap import FloodMap


class Impacts(DatabaseUser):
    """All information related to the impacts of the scenario.

    Includes methods to run the impact models or check if they has already been run.
    """

    logger = FloodAdaptLogging.getLogger("Impacts")
    name: str

    def __init__(self, scenario: Scenario):
        self.name = scenario.name
        self.scenario = scenario
        self.site_info = self.database.site
        self.models = [
            self.database.static.get_fiat_model()
        ]  # for now only FIAT adapter

    @property
    def hazard(self) -> FloodMap:
        return FloodMap(self.name)

    @property
    def socio_economic_change(self) -> SocioEconomicChange:
        return self.database.projections.get(
            self.scenario.projection
        ).socio_economic_change

    @property
    def impact_strategy(self) -> Strategy:
        return self.database.strategies.get(
            self.scenario.strategy
        ).get_impact_strategy()

    @property
    def results_path(self) -> Path:
        return db_path(
            TopLevelDir.output, object_dir=ObjectDir.scenario, obj_name=self.name
        )

    @property
    def impacts_path(self) -> Path:
        return self.results_path / "Impacts"

    @property
    def has_run(self) -> bool:
        return self.has_run_check()

    def run(self):
        """Run the impact model(s)."""
        if self.has_run:
            self.logger.info("Impacts have already been run.")
            return
        for model in self.models:
            model.run(self.scenario)

    def has_run_check(self) -> bool:
        """Check if the impact has been run.

        Returns
        -------
        bool
            _description_
        """
        checks = []
        for model in self.models:
            checks.append(model.has_run(self.scenario))
        return all(checks)



import math
from enum import Enum
from pathlib import Path
from typing import Optional, Union

from pydantic import AfterValidator, BaseModel, Field, model_validator
from tomli import load as load_toml
from typing_extensions import Annotated

from flood_adapt.objects.forcing import unit_system as us
from flood_adapt.objects.forcing.tide_gauge import TideGauge
from flood_adapt.objects.forcing.timeseries import Scstype


def ensure_ascii(s: str):
    assert s.isascii()
    return s


AsciiStr = Annotated[str, AfterValidator(ensure_ascii)]


class Cstype(str, Enum):
    """The accepted input for the variable cstype in Site."""

    projected = "projected"
    spherical = "spherical"


class SCSModel(BaseModel):
    """Class describing the accepted input for the variable scs.

    Includes the file with the non-dimensional SCS rainfall curves in the site folder and the SCS rainfall curve type.

    Attributes
    ----------
    file : str
        The path to the SCS rainfall curves file.
    type : Scstype
        The type of the SCS rainfall curve.
    """

    file: str
    type: Scstype


class RiverModel(BaseModel):
    """Model that describes the accepted input for the variable river in Site.

    Attributes
    ----------
    name : str
        The name of the river.
    description : Optional[str], default=None
        description of the river.
    mean_discharge : us.UnitfulDischarge
        The mean discharge of the river.
    x_coordinate : float
        The x coordinate of the river.
    y_coordinate : float
        The y coordinate of the river.
    """

    name: str
    description: Optional[str] = None
    mean_discharge: us.UnitfulDischarge
    x_coordinate: float
    y_coordinate: float


class ObsPointModel(BaseModel):
    """The accepted input for the variable obs_point in Site.

    obs_points is used to define output locations in the hazard model, which will be plotted in the user interface.

    Attributes
    ----------
    name : Union[int, AsciiStr]
        The name of the observation point.
    description : Optional[str], default=""
        The description of the observation point.
    ID : Optional[int], default=None
        The ID of the observation point.
    file : Optional[str], default=None
        The path to the observation point data file.
    """

    name: Union[int, AsciiStr]
    description: Optional[str] = ""
    ID: Optional[int] = (
        None  # if the observation station is also a tide gauge, this ID should be the same as for obs_station
    )
    file: Optional[str] = None  # for locally stored data
    lat: float
    lon: float


class FloodFrequencyModel(BaseModel):
    """The accepted input for the variable flood_frequency in Site."""

    flooding_threshold: us.UnitfulLength


class DemModel(BaseModel):
    """The accepted input for the variable dem in Site.

    Attributes
    ----------
    filename : str
        The path to the digital elevation model file.
    units : us.UnitTypesLength
        The units of the digital elevation model file.
    """

    filename: str
    units: us.UnitTypesLength


class FloodmapType(str, Enum):
    """The accepted input for the variable floodmap in Site."""

    water_level = "water_level"
    water_depth = "water_depth"


class DatumModel(BaseModel):
    """
    The accepted input for the variable datums in WaterlevelReferenceModel.

    Attributes
    ----------
    name : str
        The name of the vertical reference model.
    height : us.UnitfulLength
        The height of the vertical reference model relative to the main reference.
    """

    name: str
    height: us.UnitfulLength


class WaterlevelReferenceModel(BaseModel):
    """The accepted input for the variable water_level in Site.

    Waterlevels timeseries are calculated from user input, assumed to be relative to the `reference` vertical reference model.

    For plotting in the GUI, the `reference` vertical reference model is used as the main zero-reference, all values are relative to this.
    All other vertical reference models are plotted as dashed lines.

    Attributes
    ----------
    reference : str
        The name of the vertical reference model that is used as the main zero-reference.
    datums : list[DatumModel]
        The vertical reference models that are used to calculate the waterlevels timeseries.
        The datums are used to calculate the waterlevels timeseries, which are relative to the `reference` vertical reference model.
    """

    reference: str
    datums: list[DatumModel] = Field(default_factory=list)

    def get_datum(self, name: str) -> DatumModel:
        for datum in self.datums:
            if datum.name == name:
                return datum
        raise ValueError(f"Could not find datum with name {name}")

    @model_validator(mode="after")
    def main_reference_should_be_in_datums_and_eq_zero(self):
        if self.reference not in [datum.name for datum in self.datums]:
            raise ValueError(f"Reference {self.reference} not in {self.datums}")
        if not math.isclose(
            self.get_datum(self.reference).height.value, 0, abs_tol=1e-6
        ):
            raise ValueError(f"Reference {self.reference} height is not zero")
        return self

    @model_validator(mode="after")
    def all_datums_should_have_unique_names(self):
        datum_names = [datum.name for datum in self.datums]
        if len(set(datum_names)) != len(datum_names):
            raise ValueError(f"Duplicate datum names found: {datum_names}")
        return self


class CycloneTrackDatabaseModel(BaseModel):
    """The accepted input for the variable cyclone_track_database in Site.

    Attributes
    ----------
    file : str
        The path to the cyclone track database file.
    """

    file: str


class SlrScenariosModel(BaseModel):
    """The accepted input for the variable slr_scenarios.

    Attributes
    ----------
    file : str
        The path to the sea level rise scenarios file.
    relative_to_year : int
        The year to which the sea level rise scenarios are relative.
    """

    file: str
    relative_to_year: int


class FloodModel(BaseModel):
    """The accepted input for the variable overland_model and offshore_model in Site.

    Attributes
    ----------
    name : str
        The name of the directory in `static/templates/<directory>` that contains the template model files.
    reference : str
        The name of the vertical reference model that is used as the reference datum. Should be defined in water_level.datums.
    vertical_offset : Optional[us.UnitfulLength], default = None
        The vertical offset of the vertical reference model relative to the main reference.
        Given that the height of the vertical reference model is often determined by external sources,
        this vertical offset can be used to correct systematic over-/underestimation of a vertical reference model.
    """

    name: str
    reference: str

    # this used to be water_level_offset from events
    vertical_offset: Optional[us.UnitfulLength] = None


class SfincsConfigModel(BaseModel):
    """The expected variables and data types of attributes of the SfincsConfig class.

    Attributes
    ----------
    csname : str
        The name of the CS model.
    cstype : Cstype
        Cstype of the CS model. must be either "projected" or "spherical".
    version : Optional[str], default = None
        The version of the CS model. If None, the version is not specified.
    offshore_model : Optional[FloodModel], default = None
        The offshore model. If None, the offshore model is not specified.
    overland_model : FloodModel
        The overland model. This is the main model used for the simulation.
    floodmap_units : us.UnitTypesLength
        The units used for the output floodmap. Sfincs always produces in metric units, this is used to convert the floodmap to the correct units.
    save_simulation : Optional[bool], default = False
        Whether to keep or delete the simulation files after the simulation is finished and all output files are created.
        If True, the simulation files are kept. If False, the simulation files are deleted.
    """

    csname: str
    cstype: Cstype
    version: Optional[str] = None
    offshore_model: Optional[FloodModel] = None
    overland_model: FloodModel
    floodmap_units: us.UnitTypesLength
    save_simulation: Optional[bool] = False


class SfincsModel(BaseModel):
    """The expected variables and data types of attributes of the Sfincs class.

    Attributes
    ----------
    config : SfincsConfigModel
        The configuration of the Sfincs model.
    water_level : WaterlevelReferenceModel
        The collection of all datums and the main reference datum.
    dem : DemModel
        The digital elevation model.
    flood_frequency : FloodFrequencyModel, default = FloodFrequencyModel()
        The flood frequency model.
    slr : SlrScenariosModel
        Specification of the sea level rise scenarios.
    cyclone_track_database : CycloneTrackDatabaseModel, optional, default = None
        The cyclone track database model.
    scs : SCSModel, optional, default = None
        The SCS model.
    tide_gauge : TideGauge, optional, default = None
        The tide gauge model.
    river : list[RiverModel], optional, default = None
        The river model.
    obs_point : list[ObsPointModel], optional, default = None
        The observation point model.
    """

    config: SfincsConfigModel
    water_level: WaterlevelReferenceModel
    cyclone_track_database: Optional[CycloneTrackDatabaseModel] = None
    slr_scenarios: Optional[SlrScenariosModel] = None
    scs: Optional[SCSModel] = None  # optional for the US to use SCS rainfall curves
    dem: DemModel

    flood_frequency: FloodFrequencyModel = FloodFrequencyModel(
        flooding_threshold=us.UnitfulLength(value=0.0, units=us.UnitTypesLength.meters)
    )  # TODO we dont actually use this anywhere?

    tide_gauge: Optional[TideGauge] = None
    river: Optional[list[RiverModel]] = None
    obs_point: Optional[list[ObsPointModel]] = None

    @staticmethod
    def read_toml(path: Path) -> "SfincsModel":
        with open(path, mode="rb") as fp:
            toml_contents = load_toml(fp)

        return SfincsModel(**toml_contents)

    @model_validator(mode="after")
    def ensure_references_exist(self):
        datum_names = [d.name for d in self.water_level.datums]

        if self.config.overland_model.reference not in datum_names:
            raise ValueError(
                f"Could not find reference `{self.config.overland_model.reference}` in available datums: {datum_names}."
            )

        if self.config.offshore_model is not None:
            if self.config.offshore_model.reference not in datum_names:
                raise ValueError(
                    f"Could not find reference `{self.config.offshore_model.reference}` in available datums: {datum_names}."
                )

        if self.tide_gauge is not None:
            if self.tide_gauge.reference not in datum_names:
                raise ValueError(
                    f"Could not find reference `{self.tide_gauge.reference}` in available datums: {datum_names}."
                )

        return self
